# Based on: https://github.com/NVIDIA/FasterTransformer/blob/4402759e48f2340220638675f464b6ba1f79ac3c/examples/pytorch/gpt/utils/gpt_decoder.py # noqa: E501
# Copyright (c) 2022-2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from abc import abstractmethod
from pathlib import Path
from typing import List, Literal, Optional, Union
import os

import numpy as np

from . import comm
from . import profiler
from .gpt import GptInitModelParameters

from nebullvm.optional_modules.torch import torch

PathLike = Union[str, Path]


def to_numpy_dtype(maybe_str_dtype: Union[str, np.dtype]):
    assert isinstance(maybe_str_dtype, (str, np.dtype))
    if isinstance(maybe_str_dtype, str):
        try:
            dtype = {
                "fp16": np.float16,
                "float16": np.float16,
                "fp32": np.float32,
                "float32": np.float32,
            }[maybe_str_dtype]
        except KeyError:
            raise ValueError(
                f"Cannot convert to numpy data type, got {maybe_str_dtype}"
            )
    else:
        dtype = maybe_str_dtype
    return dtype


def to_torch_dtype(maybe_str_dtype: Union[str, torch.dtype]):

    if isinstance(maybe_str_dtype, torch.dtype):
        dtype = maybe_str_dtype
    else:
        try:
            dtype = {
                "bf16": torch.bfloat16,
                "fp16": torch.float16,
                "fp32": torch.float32,
                "bfloat16": torch.bfloat16,
                "float16": torch.float16,
                "float32": torch.float32,
            }[maybe_str_dtype]
        except KeyError:
            raise ValueError(
                f"Cannot convert to torch data type, got {maybe_str_dtype}"
            )
    return dtype


def load_weight_from_bin(
    checkpoint_path: PathLike,
    shape: List[int],
    weight_dtype: Union[str, np.dtype],
):
    """Load a weight from a bin file.

    # Args.
        checkpoint_path: str or Path,
            a checkpoint file path of an FT's layer weight.
        shape: list of int, the shape of weight tensor.
        weight_dtype: str or np.dtype, the data type of the stored weight.
    """
    weight_dtype = to_numpy_dtype(weight_dtype)
    return torch.from_numpy(np.fromfile(checkpoint_path, dtype=weight_dtype))


LayernormType = Literal["pre_layernorm", "post_layernorm"]


class GptLayerWeights:
    def __init__(
        self,
        num_heads: int,
        size_per_head: int,
        inter_size: int,
        num_layers: int,
        tensor_para_size: int = 1,
        pipeline_para_size: int = 1,
        has_adapters: bool = False,
        adapter_inter_size: int = 0,
        int8_mode: int = 0,
    ):

        assert num_heads % tensor_para_size == 0, (
            f"num_heads ({num_heads}) is not multiple of "
            "tensor para size ({tensor_para_size})"
        )

        self.num_heads = num_heads
        self.size_per_head = size_per_head
        self.hidden_units = num_heads * size_per_head
        self.num_layers = num_layers

        self.tensor_para_size = tensor_para_size
        self.tensor_para_rank = comm.get_tensor_para_rank()
        self.pipeline_para_size = pipeline_para_size
        self.pipeline_para_rank = comm.get_pipeline_para_rank()

        self.has_adapters = has_adapters
        self.adapter_inter_size = adapter_inter_size

        self.local_num_layers = num_layers // pipeline_para_size
        self.local_num_heads = num_heads // tensor_para_size
        self.local_hidden_units = self.local_num_heads * size_per_head
        self.local_inter_size = inter_size // tensor_para_size
        self.local_adapter_inter_size = (
            self.adapter_inter_size // tensor_para_size
        )

        self.weight_transpose_calibrate_quantize = None
        assert int8_mode in [0, 1], "Invalid int8 mode for GPT. Must be 0 or 1"
        self.int8_mode = int8_mode
        if self.int8_mode == 1:
            quant = (
                torch.ops.fastertransformer.symmetric_quantize_last_axis_of_batched_matrix  # noqa: E501
            )
            self.weight_transpose_calibrate_quantize = lambda x: quant(
                x, torch.int8
            )

        self.weights = None
        self.int8_weights = None
        self.int8_scales = None

        self.expected_weight_shapes = list()

        # pylint:disable=line-too-long
        # Transformer blocks
        self.expected_weight_shapes.extend(
            [(self.hidden_units,)] * self.local_num_layers
        )  # input layernorm weight
        self.expected_weight_shapes.extend(
            [(self.hidden_units,)] * self.local_num_layers
        )  # input layernorm bias
        self.expected_weight_shapes.extend(
            [(self.hidden_units, self.local_hidden_units * 3)]
            * self.local_num_layers
        )  # attention qkv weight
        self.expected_weight_shapes.extend(
            [(self.local_hidden_units * 3,)] * self.local_num_layers
        )  # attention qkv bias
        self.expected_weight_shapes.extend(
            [(self.local_hidden_units, self.hidden_units)]
            * self.local_num_layers
        )  # attention dense weight
        self.expected_weight_shapes.extend(
            [(self.hidden_units,)] * self.local_num_layers
        )  # attention dense bias
        self.expected_weight_shapes.extend(
            [(self.hidden_units,)] * self.local_num_layers
        )  # post attention layernorm weight
        self.expected_weight_shapes.extend(
            [(self.hidden_units,)] * self.local_num_layers
        )  # post attention layernorm bias
        self.expected_weight_shapes.extend(
            [(self.hidden_units, self.local_inter_size)]
            * self.local_num_layers
        )  # ffn_kernel1
        self.expected_weight_shapes.extend(
            [(self.local_inter_size,)] * self.local_num_layers
        )  # ffn_bias1
        self.expected_weight_shapes.extend(
            [(self.local_inter_size, self.hidden_units)]
            * self.local_num_layers
        )  # ffn_kernel2
        self.expected_weight_shapes.extend(
            [(self.hidden_units,)] * self.local_num_layers
        )  # ffn_bias2

        # Adapters
        if self.has_adapters:
            self.expected_weight_shapes.extend(
                [(self.hidden_units, self.local_adapter_inter_size)]
                * self.local_num_layers
            )  # adaptor1_kernel1
            self.expected_weight_shapes.extend(
                [(self.local_adapter_inter_size,)] * self.local_num_layers
            )  # adaptor1_bias1
            self.expected_weight_shapes.extend(
                [(self.local_adapter_inter_size, self.hidden_units)]
                * self.local_num_layers
            )  # adaptor1_kernel2
            self.expected_weight_shapes.extend(
                [(self.hidden_units,)] * self.local_num_layers
            )  # adaptor1_bias2
            self.expected_weight_shapes.extend(
                [(self.hidden_units, self.local_adapter_inter_size)]
                * self.local_num_layers
            )  # adaptor2_kernel1
            self.expected_weight_shapes.extend(
                [(self.local_adapter_inter_size,)] * self.local_num_layers
            )  # adaptor2_bias1
            self.expected_weight_shapes.extend(
                [(self.local_adapter_inter_size, self.hidden_units)]
                * self.local_num_layers
            )  # adaptor2_kernel2
            self.expected_weight_shapes.extend(
                [(self.hidden_units,)] * self.local_num_layers
            )  # adaptor2_bias2
        # pylint:enable=line-too-long

    @classmethod
    def from_config(cls, config: GptInitModelParameters):
        return cls(
            num_heads=config.head_num,
            size_per_head=config.size_per_head,
            inter_size=4 * config.head_num * config.size_per_head,
            num_layers=config.layer_num,
            tensor_para_size=config.tensor_para_size,
            pipeline_para_size=config.pipeline_para_size,
            has_adapters=config.has_adapters,
            adapter_inter_size=config.adapter_inter_size,
            int8_mode=config.int8_mode,
        )

    @property
    def dtype(self):
        return self.weights[0].dtype

    @property
    def device(self):
        return self.weights[0].device

    def _map(self, func):
        for i in range(len(self.weights)):
            if isinstance(self.weights[i], list):
                for j in range(len(self.weights[i])):
                    self.weights[i][j] = func(self.weights[i][j])
            else:
                self.weights[i] = func(self.weights[i])

    def _map_int8(self, func):
        for i in range(len(self.int8_weights)):
            if isinstance(self.int8_weights[i], list):
                for j in range(len(self.int8_weights[i])):
                    self.int8_weights[i][j] = func(self.int8_weights[i][j])

            else:
                self.int8_weights[i] = func(self.int8_weights[i])
        for i in range(len(self.int8_scales)):
            if isinstance(self.int8_scales[i], list):
                for j in range(len(self.int8_scales[i])):
                    self.int8_scales[i][j] = func(self.int8_scales[i][j])
            else:
                self.int8_scales[i] = func(self.int8_scales[i])

    def float(self):
        if self.dtype == torch.float32:
            return
        self._map(lambda x: x.float())

    def half(self):
        if self.dtype == torch.float16:
            return
        self._map(lambda x: x.half())
        if self.int8_mode == 1:
            self._map_int8(lambda w: w.half())

    def bfloat16(self):
        if self.dtype == torch.bfloat16:
            return
        self._map(lambda x: x.bfloat16())
        if self.int8_mode == 1:
            self._map_int8(lambda w: w.bfloat16())

    def cuda(self, device=None):
        self._map(lambda x: x.cuda(device))
        if self.int8_mode == 1:
            self._map_int8(lambda x: x.cuda(device))

    def to(self, device=None):
        self._map(lambda x: x.to(device))
        if self.int8_mode == 1:
            self._map_int8(lambda x: x.to(device))

    def is_valid_pp_group(self, layer, pp_rank):
        return layer // self.layers_per_device == pp_rank

    def load(
        self,
        checkpoint_path: PathLike,
        compute_dtype: torch.dtype,
        weight_dtype: Optional[Union[str, np.dtype]] = None,
        device: Optional[Union[int, str, torch.device]] = None,
    ):
        """Load checkpoint weights.

        # Args.
            checkpoint_path: str or Path,
                a checkpoint directory where FT checkpoint files locate.
            weight_dtype: str or np.dtype, the data type of stored weights.
        """

        checkpoint_path = Path(checkpoint_path)
        if not checkpoint_path.exists():
            raise FileNotFoundError(
                f"Could not find checkpoint {str(checkpoint_path)}"
            )

        weight_dtype = to_numpy_dtype(weight_dtype)
        print(
            f"Load weights from {str(checkpoint_path)} (data type: {weight_dtype}"  # noqa: E501
        )

        self.weights = list()
        self.int8_weights = list()
        self.int8_scales = list()
        torch.cuda.empty_cache()

        def _load_from_file(fname):
            quant_sub_names = [
                "attention.query_key_value.weight",
                "attention.dense.weight",
                "dense_h_to_4h.weight",
                "dense_4h_to_h.weight",
            ]
            _weight = torch.from_numpy(
                np.fromfile(checkpoint_path / fname, dtype=weight_dtype)
            )
            _weight = _weight.to(compute_dtype)
            weight_index = len(self.weights)
            expected_shape = self.expected_weight_shapes[weight_index]

            try:
                if _weight.nelement() > 0:
                    _weight = _weight.reshape(expected_shape)
            except:  # noqa: E722
                raise ValueError(
                    f"num_heads, size_per_head, vocab_size, and max_seq_len must be the same "  # noqa: E501
                    f"as the ones during training (weight: {fname} expected shape: {expected_shape}, "  # noqa: E501
                    f"got shape: {_weight.shape})."
                )

            should_quantize = any(
                sub_name in fname for sub_name in quant_sub_names
            )
            if self.int8_mode != 0 and should_quantize:
                calibrate = self.weight_transpose_calibrate_quantize
                int8_weight, int8_scales = calibrate(_weight)

                # int8 weights should appear in same order as FP weights.
                # Move to device and add to the int8 list.
                dummy_weight = torch.empty(0, dtype=compute_dtype)
                if device is not None:
                    int8_weight = int8_weight.to(device)
                    int8_scales = int8_scales.to(device)
                    dummy_weight = dummy_weight.to(device)

                self.int8_weights.append(int8_weight)
                self.int8_scales.append(int8_scales)
                self.weights.append(dummy_weight)
            else:
                if device is not None:
                    _weight = _weight.to(device)
                self.weights.append(_weight)

        # Load
        # pylint:disable=line-too-long
        layer_offset = self.local_num_layers * self.pipeline_para_rank
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.input_layernorm.weight.bin"
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.input_layernorm.bias.bin"
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.attention.query_key_value.weight.{self.tensor_para_rank}.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.attention.query_key_value.bias.{self.tensor_para_rank}.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.attention.dense.weight.{self.tensor_para_rank}.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.attention.dense.bias.bin"
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.post_attention_layernorm.weight.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.post_attention_layernorm.bias.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.mlp.dense_h_to_4h.weight.{self.tensor_para_rank}.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.mlp.dense_h_to_4h.bias.{self.tensor_para_rank}.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.mlp.dense_4h_to_h.weight.{self.tensor_para_rank}.bin"  # noqa: E501
            )
            for i in range(self.local_num_layers)
        ]
        [
            _load_from_file(
                f"model.layers.{layer_offset + i}.mlp.dense_4h_to_h.bias.bin"
            )
            for i in range(self.local_num_layers)
        ]

        if self.has_adapters:
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_attention_adapter.dense_h_to_4h.weight.{self.tensor_para_rank}.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_attention_adapter.dense_h_to_4h.bias.{self.tensor_para_rank}.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_attention_adapter.dense_4h_to_h.weight.{self.tensor_para_rank}.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_attention_adapter.dense_4h_to_h.bias.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_ffn_adapter.dense_h_to_4h.weight.{self.tensor_para_rank}.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_ffn_adapter.dense_h_to_4h.bias.{self.tensor_para_rank}.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_ffn_adapter.dense_4h_to_h.weight.{self.tensor_para_rank}.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]
            [
                _load_from_file(
                    f"model.layers.{layer_offset + i}.after_ffn_adapter.dense_4h_to_h.bias.bin"  # noqa: E501
                )
                for i in range(self.local_num_layers)
            ]

        assert len(self.weights) == len(
            self.expected_weight_shapes
        ), "Incorrect number of weights loaded"


class FtModuleBase:
    def __init__(self):
        self.weight = None

    @classmethod
    @abstractmethod
    def from_config(cls, config: GptInitModelParameters, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def _initialize_model(self, force_init=False):
        raise NotImplementedError

    @abstractmethod
    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def set_weight(self, weight: GptLayerWeights):
        old_weight_dtype = (
            self.weight.dtype if self.weight is not None else None
        )
        self.weight = weight
        if old_weight_dtype is None or old_weight_dtype != self.weight.dtype:
            self._initialize_model(force_init=True)

    @property
    def dtype(self):
        assert self.weight is not None
        return self.weight.dtype

    @property
    def device(self):
        assert self.weight is not None
        return self.weight.device

    def cuda(self, device=None):
        assert torch.cuda.is_available()
        self.weight.cuda(device)
        return self

    def to(self, device=None):
        self.weight.to(device)
        return self

    def float(self):
        self.weight.float()
        self._initialize_model(force_init=True)
        return self

    def half(self):
        self.weight.half()
        self._initialize_model(force_init=True)
        return self

    def bfloat16(self):
        self.weight.bfloat16()
        self._initialize_model(force_init=True)
        return self


class GptContextDecoder(FtModuleBase):
    def __init__(
        self,
        num_heads: int,
        size_per_head: int,
        inter_size: int,
        num_layers: int,
        tensor_para_size: int = 1,
        pipeline_para_size: int = 1,
        remove_padding: bool = True,
        shared_contexts_ratio: float = 1.0,
        layernorm_eps: float = 1e-6,
        layernorm_type: LayernormType = "pre_layernorm",
        activation_type: str = "gelu",
        has_adapters: bool = False,
        adapter_inter_size: int = 0,
        int8_mode: int = 0,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.size_per_head = size_per_head
        self.hidden_size = self.num_heads * self.size_per_head
        self.inter_size = inter_size
        self.num_layers = num_layers

        self.tensor_para_size = tensor_para_size
        self.pipeline_para_size = pipeline_para_size

        self.remove_padding = remove_padding
        self.shared_contexts_ratio = shared_contexts_ratio

        self.layernorm_eps = layernorm_eps
        self.layernorm_type = layernorm_type
        self.activation_type = activation_type
        self.has_adapters = has_adapters
        self.adapter_inter_size = adapter_inter_size

        assert int8_mode in [0, 1]
        self.int8_mode = int8_mode

        self.ft_op = None
        self.weight = None

    def __repr__(self):
        args_dict = dict(
            num_heads=self.num_heads,
            size_per_head=self.size_per_head,
            hidden_size=self.hidden_size,
            inter_size=self.inter_size,
            num_layers=self.num_layers,
            tensor_para_size=self.tensor_para_size,
            pipeline_para_size=self.pipeline_para_size,
            remove_padding=self.remove_padding,
            shared_contexts_ratio=self.shared_contexts_ratio,
            layernorm_eps=self.layernorm_eps,
            layernorm_type=self.layernorm_type,
            activation_type=self.activation_type,
            has_adapters=self.has_adapters,
            adapter_inter_size=self.adapter_inter_size,
            int8_mode=self.int8_mode,
        )
        args_str = ",\n    ".join([f"{k}: {v}" for k, v in args_dict.items()])
        return f"{self.__class__.__name__}[\n{    args_str}\n]"

    @classmethod
    def from_config(cls, config: GptInitModelParameters, **kwargs):
        return cls(
            num_heads=config.head_num,
            size_per_head=config.size_per_head,
            inter_size=4 * config.head_num * config.size_per_head,
            num_layers=config.layer_num,
            tensor_para_size=config.tensor_para_size,
            pipeline_para_size=config.pipeline_para_size,
            remove_padding=kwargs.get("remove_padding", True),
            shared_contexts_ratio=kwargs.get("shared_contexts_ratio", 1.0),
            layernorm_eps=config.layernorm_eps,
            layernorm_type=config.layernorm_type,
            activation_type=config.activation_type,
            has_adapters=config.has_adapters,
            adapter_inter_size=config.adapter_inter_size,
            int8_mode=config.int8_mode,
        )

    def _initialize_model(self, force_init=False):
        if self.weight is None:
            self.weight = GptLayerWeights(
                num_heads=self.num_heads,
                size_per_head=self.size_per_head,
                inter_size=self.inter_size,
                num_layers=self.num_layers,
                tensor_para_size=self.tensor_para_size,
                pipeline_para_size=self.pipeline_para_size,
                has_adapters=self.has_adapters,
                adapter_inter_size=self.adapter_inter_size,
                int8_mode=self.int8_mode,
            )
        if not force_init and self.ft_op is not None:
            return
        if self.ft_op is not None:
            del self.ft_op

        self.ft_op = (
            torch.classes.FasterTransformer.ParallelGptContextDecoderOp(
                self.num_heads,
                self.size_per_head,
                self.inter_size,
                self.num_layers,
                self.tensor_para_size,
                self.pipeline_para_size,
                self.layernorm_eps,
                self.layernorm_type,
                self.activation_type,
                self.has_adapters,
                self.adapter_inter_size,
                self.int8_mode,
                self.weight.weights,
                self.weight.int8_weights,
                self.weight.int8_scales,
                self.remove_padding,
            )
        )

    def forward(
        self,
        input_embeds: torch.Tensor,
        attention_mask: torch.Tensor,
        input_lengths: torch.IntTensor,
        memory_length: Optional[int] = None,
        compact_index: Optional[torch.IntTensor] = None,
        batch_to_compact_index: Optional[torch.IntTensor] = None,
        linear_bias_slopes: Optional[torch.Tensor] = None,
    ):
        """

        # Args.
            input_embeds: Tensor, (batch * beam, max_input_length, hidden_dim),
                input hidden states.
            attention_mask: Tensor, (batch * beam, max_input_length, max_input_length),
                input attention mask.
            input_lengths: (batch * beam,), input sequence lengths.
            memory_length: int, the length of memory to keep key/cache values.
            compact_index: IntTensor, (compact_batch_size,)
                The index of input sequences of a compact batch. If None, the FT op
                doesn't apply the shared context feature and as result the inference
                time may increase.
            batch_to_compact_index: IntTensor, (batch * beam,)
                The index map from the original input batch to the compact batch.
                This must be provided if compact_index is not None.
            linear_bias_slopes: (num_heads,)
                The slope per head of linear attention bias - ALiBi. If None, a base
                self attention will be performed.
        # Returns
            hidden_states: Tensor, (batch * beam, max_input_length, hidden_dim),
                decoder outputs.
            key_cache: Tensor, (num_layers, batch * beam, local_num_heads, size_per_head / x, memory_length, x), # noqa: E501
                key cache of attention of inputs.
                x = 16 / sizeof(T), memory_length = max_input_length or max_input_length + gen_length # noqa: E501
            value_cache: Tensor, (num_layers, batch * beam, local_num_heads, memory_length, hidden_dim) # noqa: E501
                value cache of attention
            last_token_hidden_states: Tensor, (batch * beam, hidden_dim)
                hidden states of the last input token.
        """
        self._initialize_model()
        # outputs: output hidden states
        (
            decoder_ouptut,
            key_cache,
            value_cache,
            last_token_hidden_states,
        ) = self.ft_op.forward(
            input_embeds,
            attention_mask,
            input_lengths,
            memory_length,
            compact_index,
            batch_to_compact_index,
            linear_bias_slopes,
        )
        return decoder_ouptut, key_cache, value_cache, last_token_hidden_states


class GptDecoder(FtModuleBase):
    def __init__(
        self,
        num_heads: int,
        size_per_head: int,
        inter_size: int,
        num_layers: int,
        tensor_para_size: int = 1,
        pipeline_para_size: int = 1,
        layernorm_eps: float = 1e-6,
        layernorm_type: LayernormType = "pre_layernorm",
        activation_type: str = "gelu",
        has_adapters: bool = False,
        adapter_inter_size: int = 0,
        int8_mode: int = 0,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.size_per_head = size_per_head
        self.hidden_size = self.num_heads * self.size_per_head
        self.inter_size = inter_size
        self.num_layers = num_layers

        self.tensor_para_size = tensor_para_size
        self.pipeline_para_size = pipeline_para_size

        self.layernorm_eps = layernorm_eps
        self.layernorm_type = layernorm_type
        self.activation_type = activation_type
        self.has_adapters = has_adapters
        self.adapter_inter_size = adapter_inter_size

        self.int8_mode = int8_mode

        self.ft_op = None
        self.weight = None

    def __repr__(self):
        args_dict = dict(
            num_heads=self.num_heads,
            size_per_head=self.size_per_head,
            hidden_size=self.hidden_size,
            inter_size=self.inter_size,
            num_layers=self.num_layers,
            tensor_para_size=self.tensor_para_size,
            pipeline_para_size=self.pipeline_para_size,
            layernorm_eps=self.layernorm_eps,
            layernorm_type=self.layernorm_type,
            activation_type=self.activation_type,
            has_adapters=self.has_adapters,
            adapter_inter_size=self.adapter_inter_size,
            int8_mode=self.int8_mode,
        )
        args_str = ",\n    ".join(
            [f"{k}: {v}" for k, v in args_dict.items()]
        )  # noqa: E501
        return f"{self.__class__.__name__}[\n    {args_str}\n]"

    @classmethod
    def from_config(cls, config: GptInitModelParameters, **kwargs):
        hidden_dim = config.head_num * config.size_per_head
        return cls(
            num_heads=config.head_num,
            size_per_head=config.size_per_head,
            inter_size=4 * hidden_dim,
            num_layers=config.layer_num,
            tensor_para_size=config.tensor_para_size,
            pipeline_para_size=config.pipeline_para_size,
            layernorm_eps=config.layernorm_eps,
            layernorm_type=config.layernorm_type,
            activation_type=config.activation_type,
            has_adapters=config.has_adapters,
            adapter_inter_size=config.adapter_inter_size,
            int8_mode=config.int8_mode,
        )

    def _initialize_model(self, force_init=False):
        if self.weight is None:
            self.weight = GptLayerWeights(
                num_heads=self.num_heads,
                size_per_head=self.size_per_head,
                inter_size=self.inter_size,
                num_layers=self.num_layers,
                tensor_para_size=self.tensor_para_size,
                pipeline_para_size=self.pipeline_para_size,
                has_adapters=self.has_adapters,
                adapter_inter_size=self.adapter_inter_size,
                int8_mode=self.int8_mode,
            )
        if not force_init and self.ft_op is not None:
            return
        if self.ft_op is not None:
            del self.ft_op
        self.ft_op = torch.classes.FasterTransformer.ParallelGptDecoderOp(
            self.num_heads,
            self.size_per_head,
            self.inter_size,
            self.num_layers,
            self.tensor_para_size,
            self.pipeline_para_size,
            self.layernorm_eps,
            self.layernorm_type,
            self.activation_type,
            self.has_adapters,
            self.adapter_inter_size,
            self.weight.int8_mode,
            self.weight.weights,
            self.weight.int8_weights,
            self.weight.int8_scales,
        )

    def forward(
        self,
        max_input_length: int,
        step: int,
        ite: int,
        input_embeds: torch.Tensor,
        sequence_lengths: torch.IntTensor,
        key_cache: torch.Tensor,
        value_cache: torch.Tensor,
        finished: torch.BoolTensor,
        total_padding_tokens: torch.IntTensor,
        masked_tokens: torch.BoolTensor,
        cache_indirection: Optional[torch.IntTensor] = None,
        linear_bias_slopes: Optional[torch.Tensor] = None,
    ):
        """

        # Args.
            max_input_length: int, maximum input context length.
            step: int, the current step index.
            ite: int, local batch iteration.
            input_embeds: Tensor, (local_batch * beam, hidden_dim),
                input hidden state to decoder.
            sequence_lengths: IntTensor, (local_batch * beam,),
                the current sequence lengths.
            key_cache: Tensor, key cache buffer.
            value_cache: Tensor, value cache buffer.
            finished: BoolTensor, (local_batch * beam,),
                whether to finish sentence generation.
            total_padding_tokens IntTensor, (local_batch * beam,),
                the number of padded tokens.
            masked_tokens: BoolTensor, (local_batch * beam, memory_length),
                a mask tensor that indicates padded tokens.
            cache_indirection: IntTensor, (local_batch * beam,),
                cache of beam positions if needed if beam > 1.
            linear_bias_slopes Tensor, (num_heads,)
                slopes head of linear position bias (ALiBi) (optional).
        # Returns
            IntTensor, (batch * beam,) output token ids.
        """

        self._initialize_model()

        outputs = self.ft_op.forward(
            max_input_length,
            step,
            ite,
            input_embeds,
            sequence_lengths,
            finished,
            total_padding_tokens,
            masked_tokens,
            key_cache,
            value_cache,
            cache_indirection,
            linear_bias_slopes,
        )
        return outputs[0]


class Gpt:
    def __init__(
        self,
        num_heads: int,
        size_per_head: int,
        num_layers: int,
        vocab_size: int,
        start_id: int,
        end_id: int,
        lib_path: PathLike,
        tensor_para_size: int = 1,
        pipeline_para_size: int = 1,
        remove_padding: bool = True,
        shared_contexts_ratio: float = 1.0,
        layernorm_eps: float = 1e-6,
        layernorm_type: LayernormType = "pre_layernorm",
        activation_type: str = "gelu",
        has_positional_encoding: bool = True,
        max_seq_len: int = 0,
        has_pre_decoder_layernorm: bool = False,
        has_post_decoder_layernorm: bool = True,
        has_adapters: bool = False,
        adapter_inter_size: int = 0,
        int8_mode: int = 0,
        inference_data_type: Optional[str] = None,
        weights_data_type: str = "fp32",
        use_fp32_to_compute_logit: bool = False,
        **kwargs,
    ):
        super().__init__()

        inference_data_type = inference_data_type or weights_data_type

        self.config = GptInitModelParameters(
            head_num=num_heads,
            size_per_head=size_per_head,
            layer_num=num_layers,
            max_seq_len=max_seq_len,
            tensor_para_size=tensor_para_size,
            vocab_size=vocab_size,
            start_id=start_id,
            end_id=end_id,
            pipeline_para_size=pipeline_para_size,
            data_type=inference_data_type,
            weights_data_type=weights_data_type,
            layernorm_eps=layernorm_eps,
            layernorm_type=layernorm_type,
            activation_type=activation_type,
            has_positional_encoding=has_positional_encoding,
            has_pre_decoder_layernorm=has_pre_decoder_layernorm,
            has_post_decoder_layernorm=has_post_decoder_layernorm,
            has_adapters=has_adapters,
            adapter_inter_size=adapter_inter_size,
            int8_mode=int8_mode,
            sparse=kwargs.get("sparse", False),
        )
        self.use_fp32_to_compute_logit = use_fp32_to_compute_logit

        self.weight = None
        self.shared_contexts_ratio = shared_contexts_ratio

        torch.classes.load_library(os.path.abspath(lib_path))

        # Embeddings to encode or decode tokens.
        hidden_dim = num_heads * size_per_head

        # Pad vocab size for FT.
        local_vocab_size = math.ceil(
            self.config.vocab_size / self.config.tensor_para_size
        )
        if self.config.data_type == "fp16":
            local_vocab_size = math.ceil(local_vocab_size / 8) * 8
        self.vocab_size_padded = (
            local_vocab_size * self.config.tensor_para_size
        )
        self.vocab_size = self.config.vocab_size

        self.decode_op = torch.classes.FasterTransformer.DynamicDecodeOp(
            self.vocab_size,
            self.vocab_size_padded,
            self.config.tensor_para_size,
            self.config.pipeline_para_size,
            torch.float,
        )

        self._parameters = {}

        def register_param(name, p):
            self._parameters[name] = p
            setattr(self, name, p)

        register_param(
            "context_decoder",
            GptContextDecoder.from_config(
                self.config,
                remove_padding=remove_padding,
                shared_contexts_ratio=shared_contexts_ratio,
                **kwargs,
            ),
        )
        register_param(
            "decoder", GptDecoder.from_config(self.config, **kwargs)
        )

        compute_dtype = to_torch_dtype(inference_data_type)

        if comm.is_pipeline_group_first():
            register_param(
                "word_embedding",
                torch.nn.Embedding(
                    self.vocab_size_padded, hidden_dim, dtype=compute_dtype
                ),
            )
            self._mask_padded_vocab_weights(self.word_embedding.weight)
            if self.config.has_positional_encoding:
                register_param(
                    "position_encoding",
                    torch.nn.Embedding(
                        self.config.max_seq_len,
                        hidden_dim,
                        dtype=compute_dtype,
                    ),
                )
            else:
                self.position_encoding = None
            if self.config.has_pre_decoder_layernorm:
                register_param(
                    "pre_decoder_layernorm",
                    torch.nn.LayerNorm(
                        hidden_dim, eps=layernorm_eps, dtype=compute_dtype
                    ),
                )
            else:
                self.pre_decoder_layernorm = None

        if comm.is_pipeline_group_last():
            if has_post_decoder_layernorm:
                register_param(
                    "post_decoder_layernorm",
                    torch.nn.LayerNorm(
                        hidden_dim, eps=layernorm_eps, dtype=compute_dtype
                    ),
                )
            else:
                self.post_decoder_layernorm = None
            self.lm_head_ctype = (
                compute_dtype
                if not self.use_fp32_to_compute_logit
                else torch.float32
            )
            register_param(
                "lm_head",
                torch.nn.Linear(
                    hidden_dim,
                    self.vocab_size_padded,
                    bias=False,
                    dtype=self.lm_head_ctype,
                ),
            )
            self._mask_padded_vocab_weights(self.lm_head.weight)

    @classmethod
    def from_config(cls, config: GptInitModelParameters, **kwargs):
        return cls(
            num_heads=config.head_num,
            size_per_head=config.size_per_head,
            num_layers=config.layer_num,
            max_seq_len=config.max_seq_len,
            tensor_para_size=config.tensor_para_size,
            vocab_size=config.vocab_size,
            start_id=config.start_id,
            end_id=config.end_id,
            pipeline_para_size=config.pipeline_para_size,
            inference_data_type=config.data_type,
            weights_data_type=config.weights_data_type,
            layernorm_eps=config.layernorm_eps,
            layernorm_type=config.layernorm_type,
            activation_type=config.activation_type,
            has_positional_encoding=config.has_positional_encoding,
            has_pre_decoder_layernorm=config.has_pre_decoder_layernorm,
            has_post_decoder_layernorm=config.has_post_decoder_layernorm,
            has_adapters=config.has_adapters,
            adapter_inter_size=config.adapter_inter_size,
            int8_mode=config.int8_mode,
            **kwargs,
        )

    def load(
        self,
        checkpoint_path: PathLike,
        inference_data_type: Optional[Union[str, torch.dtype]] = None,
        config: Optional[GptInitModelParameters] = None,
        device: Optional[Union[str, int, torch.device]] = None,
    ):

        checkpoint_path = Path(checkpoint_path)
        device = device or comm.get_device()
        config = config or self.config

        compute_dtype = to_torch_dtype(inference_data_type or self.dtype)

        self.weight = GptLayerWeights.from_config(config)
        self.weight.load(
            checkpoint_path, compute_dtype, config.weights_data_type, device
        )

        self.context_decoder.set_weight(self.weight)
        self.decoder.set_weight(self.weight)

        weight_dtype = to_numpy_dtype(config.weights_data_type)

        def _safe_load_from_bin(param: torch.nn.Parameter, fname):
            if (checkpoint_path / fname).exists():
                # np_w is 1-D array since a bin file doesn't have shape info.
                w_ = np.fromfile(checkpoint_path / fname, dtype=weight_dtype)
                param.data = (
                    torch.from_numpy(w_)
                    .reshape(param.data.shape)
                    .to(compute_dtype)
                )
            else:
                raise FileNotFoundError(f"Faile to load {fname}")

        def _safe_load_lm_head_from_bin(param, fname, ctype):
            if (checkpoint_path / fname).exists():
                shape = (
                    self.vocab_size,
                    self.config.head_num * self.config.size_per_head,
                )
                # np_w is 1-D array since a bin file doesn't have shape info.
                w_ = np.fromfile(checkpoint_path / fname, dtype=weight_dtype)
                param.data = param.data.to(ctype)
                param.data[: self.vocab_size, :] = (
                    torch.from_numpy(w_).reshape(shape).to(ctype)
                )
            else:
                print(f"Faile to load {fname}")
                torch.nn.init.normal_(param).to(compute_dtype)
            self._mask_padded_vocab_weights(param)

        # pylint:disable=line-too-long
        if comm.is_pipeline_group_first():
            _safe_load_lm_head_from_bin(
                self.word_embedding.weight, "model.wte.bin", compute_dtype
            )
            self._mask_padded_vocab_weights(self.word_embedding.weight)
            if self.position_encoding is not None:
                _safe_load_from_bin(
                    self.position_encoding.weight, "model.wpe.bin"
                )
            if self.pre_decoder_layernorm is not None:
                _safe_load_from_bin(
                    self.pre_decoder_layernorm.weight,
                    "model.pre_decoder_layernorm.weight.bin",
                )
                _safe_load_from_bin(
                    self.pre_decoder_layernorm.bias,
                    "model.pre_decoder_layernorm.bias.bin",
                )
        if comm.is_pipeline_group_last():
            if self.post_decoder_layernorm is not None:
                _safe_load_from_bin(
                    self.post_decoder_layernorm.weight,
                    "model.final_layernorm.weight.bin",
                )
                _safe_load_from_bin(
                    self.post_decoder_layernorm.bias,
                    "model.final_layernorm.bias.bin",
                )
            if (checkpoint_path / "model.lm_head.weight.bin").exists():
                _safe_load_lm_head_from_bin(
                    self.lm_head.weight,
                    "model.lm_head.weight.bin",
                    self.lm_head_ctype,
                )
            else:
                if self.use_fp32_to_compute_logit:
                    _safe_load_lm_head_from_bin(
                        self.lm_head.weight, "model.wte.bin", torch.float32
                    )
                else:
                    # In this branch we can share the pre and post
                    # decoder embeddings, but ONLY pipeline size is 1.
                    # When pipeline size > 1, these two weights will end up on
                    # different GPUs, so we must load the
                    # post decoder weight again (else case).
                    if comm.get_pipeline_para_size() == 1:
                        self.lm_head.weight = self.word_embedding.weight
                    else:
                        _safe_load_lm_head_from_bin(
                            self.lm_head.weight, "model.wte.bin", compute_dtype
                        )

        self.to(device)

    @property
    def dtype(self):
        assert self.weight is not None
        return self.weight.dtype

    @property
    def device(self):
        assert self.weight is not None
        return self.weight.device

    def cuda(self, device=None):
        assert torch.cuda.is_available()
        for name, param in self._parameters.items():
            setattr(self, name, param.cuda(device))
        return self

    def to(self, device=None):
        for name, param in self._parameters.items():
            setattr(self, name, param.to(device))
        return self

    def float(self):
        for name, param in self._parameters.items():
            setattr(self, name, param.float())
        return self

    def half(self):
        for name, param in self._parameters.items():
            setattr(self, name, param.half())
        return self

    def bfloat16(self):
        for name, param in self._parameters.items():
            setattr(self, name, param.bfloat16())
        return self

    def _mask_padded_vocab_weights(self, weight: torch.Tensor):
        assert self.vocab_size_padded >= self.vocab_size
        if self.vocab_size_padded > self.vocab_size:
            weight.data[self.vocab_size :, ...] = 0  # noqa: E203

    def generate_pad_mask(self, input_lengths, memory_length, init_step=0):
        """Generate a pad mask tensor.

        # Args.
            input_lengths: (batch_size * beam_width,), input lengths
            memory_length: the length of key/value cache memory.
            init_step: int, initial step.
        # Return
            masked_tokens: BoolTensor,
                (batch_size * beam_width, memory_length),
                True if init_step + input_length[i] <= j <
                    init_step + max_input_length,
                where i is a batch-beam index and j is a time step
                modulo by memory_length.
        """
        max_input_length = input_lengths.max()
        input_lengths = input_lengths.unsqueeze(1)
        shift = init_step % memory_length
        step_indices = torch.arange(
            init_step, init_step + memory_length, device=input_lengths.device
        )
        step_indices = (
            step_indices.roll(shift)
            .unsqueeze(0)
            .tile(input_lengths.shape[0], 1)
        )
        masked_tokens = torch.logical_and(
            step_indices >= input_lengths,
            step_indices < init_step + max_input_length,
        )
        return masked_tokens

    def get_local_batch_size(self, batch_size):
        """Get a local batch size by the same way that FT Gpt does."""
        local_batch_size = batch_size
        pp_size = self.decoder.pipeline_para_size
        if pp_size > 1:
            if local_batch_size % pp_size == 0:
                local_batch_size //= pp_size
            while local_batch_size > 1024 and local_batch_size % 2 == 0:
                local_batch_size //= 2
        return local_batch_size

    @torch.no_grad()
    def generate(
        self,
        input_token_ids: torch.IntTensor,
        input_lengths: torch.IntTensor,
        gen_length: int,
        eos_token_id: Optional[int] = None,
        local_batch_size: Optional[int] = None,
        beam_width: int = 1,
        top_k: Optional[torch.IntTensor] = None,
        top_p: Optional[torch.FloatTensor] = None,
        top_p_decay: Optional[torch.FloatTensor] = None,
        top_p_min: Optional[torch.FloatTensor] = None,
        top_p_reset_ids: Optional[torch.IntTensor] = None,
        temperature: Optional[torch.FloatTensor] = None,
        repetition_penalty: Optional[torch.FloatTensor] = None,
        presence_penalty: Optional[torch.FloatTensor] = None,
        min_length: Optional[torch.IntTensor] = None,
        len_penalty: Optional[torch.FloatTensor] = None,
        beam_search_diversity_rate: Optional[torch.FloatTensor] = None,
        stop_words_list: Optional[torch.IntTensor] = None,
        bad_words_list: Optional[torch.IntTensor] = None,
        sequence_limit_lengths: Optional[torch.IntTensor] = None,
        random_seed: Optional[torch.LongTensor] = None,
        memory_length: Optional[int] = None,
        return_output_length: bool = False,
        return_log_probs: bool = False,
    ):
        """

        # Args.
            input_token_ids: IntTensor, (batch_size, max_input_length),
                input hidden state to decoder.
            input_lengths: IntTensor, (batch_size),
                the lengths of input context sequences.
            gen_length: int, the number of tokens to generate.
            local_batch_size: int, optional, a batch size of
                local iteration. (disabled)
            eos_token_id: int, eos token id.
            beam_width: int, number of beams for beam search.
                If 1, sampling decode will be used.
            top_k: IntTensor, (batch_size,) top-k sampling.
                The number of most probable tokens to keep
                for sampling per sentence in a batcch.
            top_p: FloatTensor, (batch_size,), top-p sampling.
                The cumulative probability
                of to filter the set of most probable tokens.
            top_p_decay: FloatTensor, (batch_size,)
                The decay of top-p value for top_p sampling.
            top_p_min: FloatTensor, (batch_size,)
                The minimum top p values in top-p decaying.
            top_p_reset_ids: IntTensor, (batch_size,)
                reset ids for resetting top_p values for top p sampling
            temperature: FloatTensor, (batch_size,),
                The temperature value for smoothing the logit distribution.
            repetition_penalty: FloatTensor, (batch_size,),
                The repetition penalty.
            presence_penalty: FloatTensor, (batch_size,),
                The presence penalty, which is exclusive with
                repetition_penalty.
                Only one of repetition and presence penalties is allowed.
            min_length: IntTensor, (batch_size,),
                Minimum length for each sentences. EOS is masked if length is
                below min.
            len_penalty: FloatTensor, (batch_size,)
                The exponent of the length penalty of beam scores.
            beam_search_diversity_rate: FloatTensor, (batch_size,),
                The diversity rate of beam search.
            stop_words_list: IntTensor, (batch_size, 2, stop_words_length)
                When FT generates words in this list, it will stop the
                generation. An extension of stop id.
            bad_words_list IntTensor, (batch_size, 2, bad_words_length)
                The words in the list will never be sampled.
            sequence_limit_lengths: IntTensor, (batch_size,), The maximum
                length of a generated sequence.
            memory_length: int, the length of cache memory. If None, it will
                be max_input_length + gen_length.
        # Returns
            IntTensor, (batch_size, beam_width, max_seq_length) output
            token ids.
        """
        assert (
            self.weight is not None
        ), "Please call load() first to initialize weights."

        input_token_ids = input_token_ids.type(torch.int32).to(self.device)
        input_lengths = input_lengths.type(torch.int32).to(self.device)

        batch_size = len(input_token_ids)
        max_input_length = input_token_ids.shape[-1]
        max_seq_length = max_input_length + gen_length
        memory_length = memory_length or max_seq_length

        # TODO: Enable local batch later. We currently disable local batching due to # noqa: E501
        #   an input mismatch issue of FT's decode_op: FT's decode_op requires logits # noqa: E501
        #   of shape (batch_size, ...) but we have logits of shape (local_batch_size, ...) # noqa: E501
        #   After fixing FT's side, we will enable local batch.
        # local_batch_size = local_batch_size or self.get_local_batch_size(batch_size) # noqa: E501
        # num_local_batches, last_chunk = divmod(batch_size, local_batch_size)
        # if last_chunk > 0:
        #     num_local_batches += 1
        assert local_batch_size is None or local_batch_size == batch_size
        local_batch_size = batch_size
        num_local_batches = 1

        device = self.device

        eos_token_id = (
            eos_token_id if eos_token_id is not None else self.config.end_id
        )
        assert (
            eos_token_id is not None
        ), "eos_token-id must be specified in generation."
        eos_token_ids = eos_token_id * torch.ones(
            batch_size, dtype=torch.int32, device=device
        )
        assert repetition_penalty is None or presence_penalty is None, (
            "Found ambiguous parameters repetition_penalty and "
            "presence_penalty which are mutually exclusive. "
            "Please provide one of repetition_penalty and presence_penalty."
        )
        # Setup decoder_op prior to calling the forward function.
        self.decode_op.setup(
            batch_size,
            beam_width,
            top_k,
            top_p,
            temperature,
            repetition_penalty,
            presence_penalty,
            min_length,
            len_penalty,
            beam_search_diversity_rate,
            random_seed,
            top_p_decay,
            top_p_min,
            top_p_reset_ids,
        )

        # Prepare input and output arguments.
        if beam_width > 1:
            # Tiling for beam search.
            input_token_ids = input_token_ids.repeat(1, beam_width).view(
                batch_size * beam_width, -1
            )
            input_lengths = (
                input_lengths.view(-1, 1).repeat(1, beam_width).view(-1)
            )
            if sequence_limit_lengths is not None:
                sequence_limit_lengths = (
                    sequence_limit_lengths.view(-1, 1)
                    .repeat(1, beam_width)
                    .view(-1)
                )
            # src/tgt cache indirections.
            cache_indirection = torch.zeros(
                (2, batch_size, beam_width, memory_length),
                dtype=torch.int32,
                device=device,
            )
            parent_ids = torch.zeros(
                max_seq_length,
                batch_size * beam_width,
                dtype=torch.int32,
                device=device,
            )
        else:
            cache_indirection = None
            src_cache_indirection = None
            tgt_cache_indirection = None
            parent_ids = None

        pad_lengths = max_input_length - input_lengths
        # Since tril() doesn't support bf16 dtype,
        # we create of bool type and then cast it to dtype.
        attention_mask = (
            torch.ones(
                (max_input_length, max_input_length),
                dtype=torch.bool,
                device=device,
            )
            .tril()
            .unsqueeze(0)
            .tile(input_token_ids.shape[0], 1, 1)
            .to(self.dtype)
        )
        for b, input_length in enumerate(input_lengths):
            attention_mask[b, input_length:, ...] = 0
        masked_tokens = self.generate_pad_mask(input_lengths, memory_length)
        finished = torch.zeros_like(input_lengths).bool()
        sequence_lengths = (max_input_length - 1) * torch.ones_like(
            input_lengths
        )

        if return_log_probs or beam_width > 1:
            cum_log_probs = torch.zeros(batch_size * beam_width, device=device)
            output_log_probs = torch.zeros(
                (gen_length, batch_size * beam_width), device=device
            )
        else:
            cum_log_probs = None
            output_log_probs = None

        # Contiguous buffer for each decode_op step,
        # it will be transposed tensor for the final output.
        output_token_ids = torch.zeros(
            (max_seq_length, batch_size * beam_width),
            dtype=torch.int32,
            device=device,
        )
        output_token_ids[:max_input_length, ...] = input_token_ids.T

        if comm.is_pipeline_group_first():
            # Prepare input tensors of decoder.
            input_embeds = self.word_embedding(input_token_ids)
            if self.position_encoding is not None:
                position_ids = torch.arange(
                    0, max_input_length, dtype=torch.int, device=device
                )
                position_ids = position_ids.unsqueeze(0).view(
                    -1, max_input_length
                )
                input_embeds += self.position_encoding(position_ids)
            if self.pre_decoder_layernorm is not None:
                input_embeds = self.pre_decoder_layernorm(input_embeds)
        else:
            # Dummy input_embeds
            input_embeds = torch.empty(
                size=(
                    batch_size * beam_width,
                    max_input_length,
                    self.context_decoder.hidden_size,
                ),
                dtype=self.context_decoder.dtype,
                device=device,
            )

        use_shared_contexts = (
            (self.shared_contexts_ratio > 0.0)
            and (max_input_length >= 1)
            and (batch_size > 1)
        )
        batch_to_compact, compact_to_batch = None, None
        if use_shared_contexts:
            find_context_duplications = (
                torch.ops.fastertransformer.find_context_duplications
            )
            batch_to_compact, compact_to_batch = find_context_duplications(
                input_token_ids
            )
            use_shared_contexts = (
                compact_to_batch.shape[0]
                <= self.shared_contexts_ratio * batch_size
            )

            if not use_shared_contexts:
                batch_to_compact, compact_to_batch = None, None

        profiler.start("ft-context-decoder")
        (
            _,
            k_cache,
            v_cache,
            last_token_hidden_states,
        ) = self.context_decoder.forward(
            input_embeds=input_embeds,
            attention_mask=attention_mask,
            input_lengths=input_lengths,
            memory_length=memory_length,
            batch_to_compact_index=batch_to_compact,
            compact_index=compact_to_batch,
        )
        profiler.stop("ft-context-decoder")

        for step in range(max_input_length, max_seq_length):
            src_indir_idx = (step - max_input_length) % 2
            tgt_indir_idx = 1 - src_indir_idx

            is_generation_done = torch.tensor(
                [True], dtype=torch.bool, device=device
            )
            for ite in range(num_local_batches):
                # The indices of the current local batch-beam.
                bbidx = range(
                    ite * local_batch_size * beam_width,
                    min(
                        (ite + 1) * local_batch_size * beam_width,
                        batch_size * beam_width,
                    ),
                )
                if cache_indirection is not None:
                    bidx = range(
                        ite * local_batch_size,
                        min((ite + 1) * local_batch_size, batch_size),
                    )
                    src_cache_indirection = cache_indirection[
                        src_indir_idx, bidx, ...
                    ]
                    tgt_cache_indirection = cache_indirection[
                        tgt_indir_idx, bidx, ...
                    ]

                if step == max_input_length:
                    hidden_states = last_token_hidden_states[bbidx, ...]
                else:
                    if comm.is_pipeline_group_first():
                        input_embeds = self.word_embedding(
                            output_token_ids[step - 1, bbidx]
                        )
                        if self.position_encoding is not None:
                            position_ids = (step - 1) * torch.ones_like(
                                pad_lengths[bbidx]
                            )
                            input_embeds += self.position_encoding(
                                position_ids
                            )
                        if self.pre_decoder_layernorm is not None:
                            input_embeds = self.pre_decoder_layernorm(
                                input_embeds
                            )
                    else:
                        # Dummy input_imbeds
                        input_embeds = torch.empty(
                            size=(len(bbidx), self.decoder.hidden_size),
                            dtype=self.decoder.dtype,
                            device=device,
                        )

                    profiler.start("ft-decoder")
                    hidden_states = self.decoder.forward(
                        max_input_length=max_input_length,
                        step=step,
                        ite=ite,
                        input_embeds=input_embeds,
                        sequence_lengths=sequence_lengths[bbidx],
                        key_cache=k_cache,
                        value_cache=v_cache,
                        finished=finished[bbidx],
                        total_padding_tokens=pad_lengths[bbidx],
                        cache_indirection=src_cache_indirection,
                        masked_tokens=masked_tokens[bbidx, ...],
                    )
                    profiler.stop("ft-decoder")

                if comm.is_pipeline_group_last():
                    if self.post_decoder_layernorm is not None:
                        hidden_states = self.post_decoder_layernorm(
                            hidden_states
                        )

                    # We use logits of fp32 type to avoid overflow issue.
                    if self.use_fp32_to_compute_logit:
                        # The FT GPT op internally uses FP32 compute type
                        # for matrix multiplication.
                        # This will produce the same result with the
                        # end-to-end FT's GPT op.
                        logits = torch.nn.functional.linear(
                            hidden_states.float(), self.lm_head.weight
                        )
                    else:
                        logits = self.lm_head(hidden_states).float()

                    profiler.start("ft-decode")
                    should_stop = self.decode_op.forward(
                        logits.view(batch_size, beam_width, -1),
                        step,
                        max_input_length,
                        ite,
                        local_batch_size,
                        eos_token_ids,
                        top_k,
                        top_p,
                        temperature,
                        repetition_penalty,
                        presence_penalty,
                        min_length,
                        len_penalty,
                        beam_search_diversity_rate,
                        top_p_decay,
                        top_p_min,
                        top_p_reset_ids,
                        None,
                        input_lengths,
                        sequence_limit_lengths,
                        stop_words_list,
                        bad_words_list,
                        src_cache_indirection,
                        output_token_ids.view(-1, batch_size, beam_width),
                        finished,
                        sequence_lengths,
                        cum_log_probs,
                        output_log_probs,
                        parent_ids,
                        tgt_cache_indirection,
                    )
                    profiler.stop("ft-decode")
                    is_generation_done &= should_stop

            # Broadcast from the last pipeline node if needed.
            profiler.start("ft-bcast")
            tensors_to_bcast = [
                output_token_ids[step, ...],
                finished,
                sequence_lengths,
                is_generation_done,
            ]
            if beam_width > 1:
                tensors_to_bcast.append(tgt_cache_indirection)
            self.decode_op.broadcast_from_last_pipeline(tensors_to_bcast)
            profiler.stop("ft-bcast")

            if is_generation_done or finished.all():
                break

        # Transpose (L, batch, beam) -> (batch, beam, L)
        output_token_ids = output_token_ids.view(
            -1, batch_size, beam_width
        ).permute(1, 2, 0)

        # Increase sequence_length by 1 because the sequence length of time step t is t - 1. # noqa: E501
        sequence_lengths += 1

        # Outputs
        output_dict = dict(output_token_ids=output_token_ids)
        if return_output_length:
            output_dict["output_lengths"] = sequence_lengths
        if return_log_probs:
            output_dict["cum_log_probs"] = cum_log_probs
            output_dict["output_log_probs"] = output_log_probs
        return output_dict



from contextlib import nullcontext
from pathlib import Path

from loguru import logger

from nebullvm.config import ONNX_OPSET_VERSION
from nebullvm.core.models import ModelParams, Device, DeviceType, DataType
from nebullvm.optional_modules.torch import torch, Module
from nebullvm.tools.data import DataManager
from nebullvm.tools.pytorch import (
    create_model_inputs_torch,
)


@torch.inference_mode()
def convert_torch_to_onnx(
    torch_model: Module,
    input_data: DataManager,
    model_params: ModelParams,
    output_file_path: Path,
    device: Device,
):
    """Function importing a custom model in pytorch and converting it in ONNX

    Args:
        torch_model (Module): Pytorch model.
        input_data (DataManager): Custom data provided by user to be
        used as input for the converter.
        model_params (ModelParams): Model Parameters as input sizes and
            dynamic axis information.
        output_file_path (str or Path): Path where storing the output
            ONNX file.
        device (Device): Device where the model will be run.
    """

    if input_data is not None:
        input_tensors = list(input_data.get_list(1)[0])
    else:
        input_tensors = create_model_inputs_torch(model_params.input_infos)

    output_sizes = model_params.output_sizes
    output_types = model_params.output_types

    input_names = [f"input_{i}" for i in range(len(input_tensors))]
    output_names = [f"output_{i}" for i in range(len(output_sizes))]
    dynamic_info = model_params.dynamic_info

    if dynamic_info is not None:
        # This check is needed to enable backward compatibility with
        # previous versions of nebullvm
        if isinstance(list(dynamic_info.inputs[0].values())[0], str):
            onnx_format_inputs = dynamic_info.inputs
        else:
            onnx_format_inputs = [
                {k: v["name"] for (k, v) in d.items()}
                for d in dynamic_info.inputs
            ]

        assert len(dynamic_info.outputs) == len(output_names), (
            f"The number of dynamic outputs provided in the dynamic info "
            f"dict ({len(dynamic_info.outputs)}) is not equal to the number "
            f"of outputs of the model ({len(output_names)}), Detected model "
            f"output shapes are: {output_sizes} "
        )

        dynamic_info = {
            name: dynamic_dict
            for name, dynamic_dict in zip(
                input_names + output_names,
                onnx_format_inputs + dynamic_info.outputs,
            )
        }

    try:
        # try conversion with model on cpu
        if device.type is DeviceType.GPU:
            input_tensors = [x.cpu() for x in input_tensors]
            torch_model.cpu()

        torch.onnx.export(
            torch_model,  # model being run
            tuple(
                input_tensors
            ),  # model input (or a tuple for multiple inputs)
            str(output_file_path),
            # where to save the model (can be a file or file-like object)
            export_params=True,
            # store the trained parameter weights inside the model file
            opset_version=ONNX_OPSET_VERSION,
            # the ONNX version to export the model to
            do_constant_folding=True,
            # whether to execute constant folding for optimization
            input_names=input_names,
            # the model's input names
            output_names=output_names,
            # the model's output names
            dynamic_axes=dynamic_info,
        )

        # Put again model on gpu
        if device.type is DeviceType.GPU:
            torch_model.to(device.to_torch_format())

        return output_file_path
    except Exception:
        # try conversion with model on gpu
        if device.type is DeviceType.GPU:
            input_tensors = [
                x.to(device.to_torch_format()) for x in input_tensors
            ]
            torch_model.to(device.to_torch_format())

            try:
                with torch.autocast("cuda") if output_types[
                    0
                ] is DataType.FLOAT16 else nullcontext():
                    torch.onnx.export(
                        torch_model,  # model being run
                        tuple(
                            input_tensors
                        ),  # model input (or a tuple for multiple inputs)
                        str(output_file_path),
                        # where to save the model
                        # (can be a file or file-like object)
                        export_params=True,
                        # store the trained parameter weights inside the model
                        opset_version=ONNX_OPSET_VERSION,
                        # the ONNX version to export the model to
                        do_constant_folding=True,
                        # whether to execute constant folding for optimization
                        input_names=input_names,
                        # the model's input names
                        output_names=output_names,
                        # the model's output names
                        dynamic_axes=dynamic_info,
                    )

                return output_file_path
            except Exception:
                logger.warning(
                    "Exception raised during conversion from torch"
                    " to onnx model. ONNX pipeline will be unavailable."
                )
                return None
        else:
            logger.warning(
                "Exception raised during conversion from torch"
                " to onnx model. ONNX pipeline will be unavailable."
            )
            return None



import os
import pickle
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Tuple, Union

from nebullvm.core.models import Device, DeviceType, ModelParams
from nebullvm.operations.inference_learners.base import (
    PytorchBaseInferenceLearner,
    LearnerMetadata,
)
from nebullvm.optional_modules.torch import (
    torch,
)
from nebullvm.tools.transformations import MultiStageTransformation


class TorchXLAInferenceLearner(PytorchBaseInferenceLearner):
    MODEL_NAME = "model_scripted.pt"
    name = "TorchXLA"

    def __init__(self, torch_model: torch.nn.Module, device: Device, **kwargs):
        super().__init__(**kwargs)
        self.model = torch_model.eval()
        if device.type is DeviceType.TPU:
            self.model.to(device.to_torch_format())
        self.device = device
        self._is_gpu_ready = self.device.type is DeviceType.TPU

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        if self.device.type is DeviceType.TPU and not self._is_gpu_ready:
            self.set_model_on_gpu()
        if self.device.type is DeviceType.TPU:
            input_tensors = (
                t.to(self.device.to_torch_format()) for t in input_tensors
            )
        with torch.no_grad():
            res = self.model(*input_tensors)
            if not isinstance(res, tuple):
                return (res,)
            return tuple(out for out in res)

    def get_size(self):
        try:
            if hasattr(self.model, "core_model"):
                return len(pickle.dumps(self.model.core_model, -1))
            else:
                # Normal torch model
                return len(pickle.dumps(self.model, -1))
        except RuntimeError:
            with TemporaryDirectory() as tmp_dir:
                self.save(tmp_dir)
                return sum(
                    os.path.getsize(Path(tmp_dir) / f)
                    for f in os.listdir(Path(tmp_dir))
                    if os.path.isfile(Path(tmp_dir) / f)
                )

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        path.mkdir(exist_ok=True)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)
        self.model.cpu()
        torch.save(self.model, path / self.MODEL_NAME)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        model = torch.load(path / cls.MODEL_NAME)
        metadata = LearnerMetadata.read(path)
        device = Device.from_str(metadata.device)
        model.to(device.to_torch_format())
        return cls(
            torch_model=model,
            network_parameters=ModelParams(**metadata.network_parameters),
            input_tfms=MultiStageTransformation.from_dict(metadata.input_tfms)
            if metadata.input_tfms is not None
            else None,
            device=device,
        )



import os
import platform

import cpuinfo
import psutil

from nebullvm.core.models import HardwareSetup, Device, DeviceType
from nebullvm.optional_modules.torch_xla import xm
from nebullvm.optional_modules.utils import (
    torch_is_available,
    tensorflow_is_available,
)
from nebullvm.tools.pytorch import torch_get_device_name
from nebullvm.tools.tf import tensorflow_get_gpu_name
from nebullvm.tools.utils import (
    gpu_is_available,
    tpu_is_available,
    neuron_is_available,
)


def get_hw_setup(device: Device = None) -> HardwareSetup:
    accelerator = None
    if (
        device is not None and device.type is DeviceType.GPU
    ) or gpu_is_available():
        accelerator = _get_gpu_name()
    elif (
        device is not None and device.type is DeviceType.TPU
    ) or tpu_is_available():
        accelerator = _get_tpu_device_name()
    elif (
        device is not None and device.type is DeviceType.NEURON
    ) or neuron_is_available():
        accelerator = _get_neuron_device_name()
    return HardwareSetup(
        cpu=cpuinfo.get_cpu_info()["brand_raw"],
        operating_system=platform.system(),
        memory_gb=round(psutil.virtual_memory().total * 1e-9, 2),
        accelerator=accelerator,
    )


def _get_gpu_name() -> str:
    if torch_is_available():
        name = torch_get_device_name()
    elif tensorflow_is_available():
        name = tensorflow_get_gpu_name()
    else:
        name = "Unknown"
    return name


def _get_neuron_device_name() -> str:
    output = os.popen("lshw -businfo").read()
    neuron_name = "Unknown Neuron"
    for line in output.splitlines():
        if "neuron" in line.lower():
            words = line.split(" ")
            if len(words) > 2:
                neuron_name = " ".join(words[-2:])
                break
    return neuron_name


def _get_tpu_device_name() -> str:
    return xm.xla_device_hw(xm.xla_device())



import abc
from tempfile import TemporaryDirectory
from typing import Any, Callable, Dict, List, Tuple, Type, Union

from nebullvm.config import ACTIVATION_METRIC_DROP_THS
from nebullvm.core.models import (
    OptimizedModel,
    OptimizationTime,
    ModelParams,
    ModelCompiler,
    ModelCompressor,
    DeepLearningFramework,
    DeviceType,
    QuantizationType,
)
from nebullvm.operations.base import Operation
from nebullvm.operations.inference_learners.base import (
    BuildInferenceLearner,
)
from nebullvm.operations.inference_learners.builders import (
    DeepSparseBuildInferenceLearner,
    FasterTransformerBuildInferenceLearner,
    IntelNeuralCompressorBuildInferenceLearner,
    ONNXApacheTVMBuildInferenceLearner,
    ONNXBuildInferenceLearner,
    ONNXTensorRTBuildInferenceLearner,
    OpenVINOBuildInferenceLearner,
    PyTorchApacheTVMBuildInferenceLearner,
    PyTorchTensorRTBuildInferenceLearner,
    TensorflowBuildInferenceLearner,
    TFLiteBuildInferenceLearner,
    TorchNeuronBuildInferenceLearner,
    TorchXLABuildInferenceLearner,
    TorchDynamoBuildInferenceLearner,
    TorchScriptBuildInferenceLearner,
)
from nebullvm.operations.measures.measures import MetricDropMeasure
from nebullvm.operations.measures.utils import (
    compute_optimized_running_time,
    compute_relative_difference,
)
from nebullvm.operations.optimizations.compilers.base import Compiler
from nebullvm.operations.optimizations.compilers.deepsparse import (
    DeepSparseCompiler,
)
from nebullvm.operations.optimizations.compilers.faster_transformer import (
    FasterTransformerCompiler,
)
from nebullvm.operations.optimizations.compilers.intel_neural_compressor import (  # noqa: E501
    IntelNeuralCompressorCompiler,
)
from nebullvm.operations.optimizations.compilers.onnxruntime import (
    ONNXCompiler,
)
from nebullvm.operations.optimizations.compilers.openvino import (
    OpenVINOCompiler,
)
from nebullvm.operations.optimizations.compilers.tensor_rt import (
    ONNXTensorRTCompiler,
    PyTorchTensorRTCompiler,
)
from nebullvm.operations.optimizations.compilers.tensorflow import (
    TensorflowBackendCompiler,
    TFLiteBackendCompiler,
)
from nebullvm.operations.optimizations.compilers.torch_dynamo import (
    TorchDynamoCompiler,
)
from nebullvm.operations.optimizations.compilers.torch_neuron import (
    TorchNeuronCompiler,
)
from nebullvm.operations.optimizations.compilers.torch_xla import (
    TorchXLACompiler,
)
from nebullvm.operations.optimizations.compilers.torchscript import (
    TorchScriptCompiler,
)
from nebullvm.operations.optimizations.compilers.tvm import (
    ONNXApacheTVMCompiler,
    PyTorchApacheTVMCompiler,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation
from nebullvm.tools.utils import get_throughput


class Optimizer(Operation, abc.ABC):
    def __init__(self):
        super().__init__()
        self.optimized_models = []
        self.source_dl_framework = None
        self.pipeline_dl_framework = None
        self.compiler_ops = {}
        self.build_inference_learner_ops = {}
        self.validity_check_op = MetricDropMeasure()

    def execute(
        self,
        model: Any,
        input_data: DataManager,
        optimization_time: OptimizationTime,
        metric_drop_ths: float,
        metric: Callable,
        model_params: ModelParams,
        model_outputs: List[Tuple[Any, ...]],
        ignore_compilers: List[ModelCompiler],
        ignore_compressors: List[ModelCompressor],
        source_dl_framework: DeepLearningFramework,
        is_diffusion: bool = False,
    ) -> List[OptimizedModel]:
        self.source_dl_framework = source_dl_framework

        # TODO: implement and select compressors from hardware

        compilers = self._select_compilers_from_hardware()

        remove_compiler_list = []
        add_compiler_list = []
        for compiler in ignore_compilers:
            if compiler in MULTI_FRAMEWORK_COMPILERS:
                add_compiler_list += MULTI_FRAMEWORK_COMPILERS[compiler]
                remove_compiler_list.append(compiler)

        for c in remove_compiler_list:
            ignore_compilers.remove(c)

        ignore_compilers += add_compiler_list

        (
            self.compiler_ops,
            self.build_inference_learner_ops,
        ) = self._load_compilers(
            ignore_compilers=ignore_compilers,
            compilers=compilers,
        )
        self._optimize(
            model=model,
            input_data=input_data,
            optimization_time=optimization_time,
            metric_drop_ths=metric_drop_ths,
            metric=metric,
            model_params=model_params,
            model_outputs=model_outputs,
            ignore_compilers=ignore_compilers,
            is_diffusion=is_diffusion,
        )

        return self.optimized_models

    @abc.abstractmethod
    def _select_compilers_from_hardware(self):
        raise NotImplementedError()

    @staticmethod
    def _load_compilers(
        ignore_compilers: List[ModelCompiler],
        compilers: List[ModelCompiler],
    ):
        compiler_ops = {
            compiler: COMPILER_TO_OPTIMIZER_MAP[compiler]()
            for compiler in compilers
            if compiler not in ignore_compilers
            and compiler in COMPILER_TO_OPTIMIZER_MAP
        }
        build_inference_learner_ops = {
            compiler: COMPILER_TO_INFERENCE_LEARNER_MAP[compiler]()
            for compiler in compilers
            if compiler not in ignore_compilers
            and compiler in COMPILER_TO_OPTIMIZER_MAP
        }

        return compiler_ops, build_inference_learner_ops

    def free_model_gpu(self, model: Any):
        # Free gpu memory
        if self.device.type is DeviceType.GPU:
            try:
                model.cpu()
            except Exception:
                pass
            try:
                with torch.cuda.device(self.device.to_torch_format()):
                    torch.cuda.empty_cache()
            except Exception:
                pass

    def _optimize(
        self,
        model: Union[torch.nn.Module, tf.Module, str],
        input_data: DataManager,
        optimization_time: OptimizationTime,
        metric_drop_ths: float,
        metric: Callable,
        model_params: ModelParams,
        model_outputs: List[Tuple[Any, ...]],
        ignore_compilers: List[ModelCompiler],
        is_diffusion: bool = False,
    ):

        if metric_drop_ths is not None:
            q_types = [
                None,
            ]
            if metric_drop_ths > 0:
                q_types.append(QuantizationType.HALF)
            if metric_drop_ths > ACTIVATION_METRIC_DROP_THS:
                q_types.append(QuantizationType.DYNAMIC)
                if input_data is not None:
                    q_types.append(QuantizationType.STATIC)
        else:
            q_types = [None]

        optimization_info = []
        for compiler, compiler_op, build_inference_learner_op in zip(
            self.compiler_ops.keys(),
            self.compiler_ops.values(),
            self.build_inference_learner_ops.values(),
        ):
            for q_type in q_types:
                input_tfms = MultiStageTransformation([])

                self.free_model_gpu(model)

                with TemporaryDirectory() as tmp_dir:
                    try:
                        compiler_op.to(self.device).execute(
                            model=model,
                            input_data=input_data,
                            model_params=model_params,
                            metric_drop_ths=metric_drop_ths
                            if q_type is not None
                            else None,
                            quantization_type=q_type,
                            input_tfms=input_tfms,
                            onnx_output_path=tmp_dir,
                            is_diffusion=is_diffusion,
                        )

                        compiled_model = compiler_op.get_result()
                        if compiled_model is not None:
                            build_inference_learner_op.to(self.device).execute(
                                model=compiled_model,
                                model_orig=compiler_op.model_orig
                                if hasattr(compiler_op, "model_orig")
                                else None,
                                model_params=model_params,
                                input_tfms=input_tfms,
                                source_dl_framework=self.source_dl_framework,
                                quantization_type=q_type,
                            )
                            inference_learner = (
                                build_inference_learner_op.get_result()
                            )

                            if inference_learner is not None:
                                test_input_data, ys = input_data.get_split(
                                    "test"
                                ).get_list(with_ys=True)

                                self.validity_check_op.execute(
                                    inference_learner,
                                    test_input_data,
                                    model_outputs,
                                    metric_drop_ths,
                                    metric_func=metric
                                    if q_type is not None
                                    else compute_relative_difference,
                                    ys=ys,
                                )

                                if self.validity_check_op.valid:
                                    latency = compute_optimized_running_time(
                                        inference_learner, input_data
                                    )
                                    self.logger.info(
                                        f"Optimized model latency: {latency} "
                                        f"sec/iter"
                                    )

                                    if (
                                        compiler not in ignore_compilers
                                        and optimization_time
                                        is OptimizationTime.CONSTRAINED
                                    ):
                                        ignore_compilers.append(compiler)

                                    self.optimized_models.append(
                                        OptimizedModel(
                                            inference_learner=inference_learner,  # noqa: E501
                                            metric_drop=self.validity_check_op.measure_result,  # noqa: E501
                                            compiler=compiler,
                                            technique=q_type.name
                                            if q_type is not None
                                            else "None",
                                            latency_seconds=latency,
                                            throughput=get_throughput(
                                                latency,
                                                # Normal models have batch
                                                # size B, diffusion models
                                                # have batch size 2B
                                                model_params.batch_size
                                                if not is_diffusion
                                                else model_params.batch_size
                                                / 2,
                                            ),
                                            size_mb=inference_learner.get_size()  # noqa: E501
                                            / 1e6,
                                        )
                                    )

                                    opt_info_dict = {
                                        "compiler": f"{self.pipeline_dl_framework.value}_{compiler.value}",  # noqa: E501
                                        "technique": q_type.value
                                        if q_type
                                        else "none",
                                        "latency": latency,
                                    }
                                    if (
                                        metric_drop_ths is not None
                                        and q_type is not None
                                    ):
                                        opt_info_dict[
                                            "metric_loss"
                                        ] = (
                                            self.validity_check_op.measure_result  # noqa: E501
                                        )
                                        opt_info_dict[
                                            "metric"
                                        ] = metric.__name__
                                    optimization_info.append(opt_info_dict)
                                else:
                                    self.logger.warning(
                                        "The optimized model will be "
                                        "discarded due to poor results "
                                        "obtained with the given metric."
                                    )

                                if self.device.type in [
                                    DeviceType.GPU,
                                    DeviceType.TPU,
                                ]:
                                    inference_learner.free_gpu_memory()
                    except Exception as ex:
                        self.logger.warning(
                            f"Optimization failed with "
                            f"{self.pipeline_dl_framework} "
                            f"interface of {compiler}. Got error {ex}. "
                            f"If possible the compilation will be re-scheduled"
                            f" with another interface. Please consult the "
                            f"documentation for further info or open an issue "
                            f"on GitHub for receiving assistance."
                        )
                        optimization_info.append(
                            {
                                "compiler": compiler.value,
                                "technique": q_type.value
                                if q_type
                                else "none",
                                "latency": -1,
                            }
                        )
        if self.feedback_collector is not None:
            self.feedback_collector.store_info(
                key="optimizations",
                value=optimization_info,
            )


MULTI_FRAMEWORK_COMPILERS = {
    ModelCompiler.TENSOR_RT: [
        ModelCompiler.TENSOR_RT_TORCH,
        ModelCompiler.TENSOR_RT_ONNX,
    ],
    ModelCompiler.APACHE_TVM: [
        ModelCompiler.APACHE_TVM_TORCH,
        ModelCompiler.APACHE_TVM_ONNX,
    ],
}

COMPILER_TO_OPTIMIZER_MAP: Dict[ModelCompiler, Type[Compiler]] = {
    ModelCompiler.TORCHSCRIPT: TorchScriptCompiler,
    ModelCompiler.DEEPSPARSE: DeepSparseCompiler,
    ModelCompiler.INTEL_NEURAL_COMPRESSOR: IntelNeuralCompressorCompiler,
    ModelCompiler.TENSOR_RT_TORCH: PyTorchTensorRTCompiler,
    ModelCompiler.TENSOR_RT_ONNX: ONNXTensorRTCompiler,
    ModelCompiler.APACHE_TVM_TORCH: PyTorchApacheTVMCompiler,
    ModelCompiler.APACHE_TVM_ONNX: ONNXApacheTVMCompiler,
    ModelCompiler.ONNX_RUNTIME: ONNXCompiler,
    ModelCompiler.OPENVINO: OpenVINOCompiler,
    ModelCompiler.TFLITE: TFLiteBackendCompiler,
    ModelCompiler.XLA: TensorflowBackendCompiler,
    ModelCompiler.TORCH_NEURON: TorchNeuronCompiler,
    ModelCompiler.TORCH_XLA: TorchXLACompiler,
    ModelCompiler.TORCH_DYNAMO: TorchDynamoCompiler,
    ModelCompiler.FASTER_TRANSFORMER: FasterTransformerCompiler,
}

COMPILER_TO_INFERENCE_LEARNER_MAP: Dict[
    ModelCompiler, Type[BuildInferenceLearner]
] = {
    ModelCompiler.TORCHSCRIPT: TorchScriptBuildInferenceLearner,
    ModelCompiler.DEEPSPARSE: DeepSparseBuildInferenceLearner,
    ModelCompiler.INTEL_NEURAL_COMPRESSOR: IntelNeuralCompressorBuildInferenceLearner,  # noqa: E501
    ModelCompiler.TENSOR_RT_TORCH: PyTorchTensorRTBuildInferenceLearner,
    ModelCompiler.TENSOR_RT_ONNX: ONNXTensorRTBuildInferenceLearner,
    ModelCompiler.APACHE_TVM_TORCH: PyTorchApacheTVMBuildInferenceLearner,
    ModelCompiler.APACHE_TVM_ONNX: ONNXApacheTVMBuildInferenceLearner,
    ModelCompiler.ONNX_RUNTIME: ONNXBuildInferenceLearner,
    ModelCompiler.OPENVINO: OpenVINOBuildInferenceLearner,
    ModelCompiler.TFLITE: TFLiteBuildInferenceLearner,
    ModelCompiler.XLA: TensorflowBuildInferenceLearner,
    ModelCompiler.TORCH_NEURON: TorchNeuronBuildInferenceLearner,
    ModelCompiler.TORCH_XLA: TorchXLABuildInferenceLearner,
    ModelCompiler.TORCH_DYNAMO: TorchDynamoBuildInferenceLearner,
    ModelCompiler.FASTER_TRANSFORMER: FasterTransformerBuildInferenceLearner,
}



from pathlib import Path

import nebullvm
from nebullvm.core.models import Device, ModelCompiler, DeviceType


def onnxruntime_is_available() -> bool:
    try:
        import onnxruntime  # noqa F401

        return True
    except ImportError:
        return False


def tvm_is_available() -> bool:
    try:
        import tvm  # noqa F401
        from tvm.runtime import Module  # noqa F401

        return True
    except ImportError:
        return False


def bladedisc_is_available() -> bool:
    try:
        import torch_blade  # noqa F401

        return True
    except ImportError:
        return False


def tensorrt_is_available() -> bool:
    try:
        import polygraphy  # noqa F401
        import tensorrt  # noqa F401

        return True
    except ImportError:
        return False


def torch_tensorrt_is_available() -> bool:
    try:
        import torch_tensorrt  # noqa F401

        return True
    except ImportError:
        return False


def openvino_is_available() -> bool:
    try:
        from openvino.runtime import Core  # noqa F401
    except ImportError:
        return False
    else:
        return True


def deepsparse_is_available() -> bool:
    try:
        import deepsparse  # noqa F401
    except ImportError:
        return False
    else:
        return True


def intel_neural_compressor_is_available() -> bool:
    try:
        import neural_compressor  # noqa F401
    except ImportError:
        return False
    else:
        return True


def torch_xla_is_available():
    try:
        import torch_xla  # noqa F401

        return True
    except ImportError:
        return False


def torch_neuron_is_available():
    try:
        import torch_neuron  # noqa F401

        return True
    except ImportError:
        return False


def get_faster_transformer_repo_path() -> Path:
    return Path(nebullvm.__file__).parent.joinpath("FasterTransformer")


def faster_transformer_is_available() -> bool:
    return (
        get_faster_transformer_repo_path()
        .parent.joinpath("FasterTransformer_build_success")
        .exists()
    )


def select_compilers_from_hardware_onnx(device: Device):
    from nebullvm.optional_modules.utils import onnx_is_available

    compilers = []
    if onnx_is_available():
        if onnxruntime_is_available():
            compilers.append(ModelCompiler.ONNX_RUNTIME)
        if tvm_is_available():
            compilers.append(ModelCompiler.APACHE_TVM)
        if device.type is DeviceType.GPU and tensorrt_is_available():
            compilers.append(ModelCompiler.TENSOR_RT)
        if device.type is DeviceType.CPU and openvino_is_available():
            compilers.append(ModelCompiler.OPENVINO)
    return compilers


def select_compilers_from_hardware_torch(device: Device):
    from nebullvm.optional_modules.utils import torch_is_available

    compilers = []
    if torch_is_available():
        compilers.append(ModelCompiler.TORCHSCRIPT)
        if tvm_is_available():
            compilers.append(ModelCompiler.APACHE_TVM)
        if bladedisc_is_available():
            compilers.append(ModelCompiler.BLADEDISC)
        if torch_neuron_is_available():
            compilers.append(ModelCompiler.TORCH_NEURON)

        if device.type is DeviceType.CPU:
            if deepsparse_is_available():
                compilers.append(ModelCompiler.DEEPSPARSE)
            if intel_neural_compressor_is_available():
                compilers.append(ModelCompiler.INTEL_NEURAL_COMPRESSOR)
        elif device.type is DeviceType.GPU:
            if torch_tensorrt_is_available:
                compilers.append(ModelCompiler.TENSOR_RT)
    return compilers


def select_compilers_from_hardware_tensorflow():
    from nebullvm.optional_modules.utils import tensorflow_is_available

    compilers = []
    if tensorflow_is_available():
        compilers.append(ModelCompiler.XLA)
        compilers.append(ModelCompiler.TFLITE)
    return compilers



from typing import List, Tuple

from nebullvm.core.models import QuantizationType, ModelParams, DeviceType
from nebullvm.operations.optimizations.compilers.base import Compiler
from nebullvm.operations.optimizations.compilers.quantizations.utils import (
    check_quantization,
)
from nebullvm.optional_modules.torch import (
    torch,
    symbolic_trace,
)
from nebullvm.optional_modules.torch_neuron import torch_neuron
from nebullvm.tools.data import DataManager
from nebullvm.tools.transformations import MultiStageTransformation


class TorchNeuronCompiler(Compiler):
    supported_ops = {
        "cpu": [],
        "gpu": [],
        "neuron": [None, QuantizationType.HALF],
    }

    @staticmethod
    def _check_dynamic_shape(network_parameters: ModelParams) -> bool:
        """Handles case when model inputs have dynamic shapes.
        For now TorchNeuron only supports dynamic shape for the
        batch dimension.

        Args:
            network_parameters (ModelParams): The model parameters.

        Returns:
            bool: True if the model has dynamic batch size, False otherwise.
        """
        if network_parameters.dynamic_info is None:
            return False

        for i, input_shape in enumerate(
            network_parameters.dynamic_info.inputs
        ):
            if len(input_shape) > 1 or (
                len(input_shape) == 1 and input_shape.get(0) is None
            ):
                raise ValueError(
                    f"TorchNeuronCompiler only supports dynamic shapes for "
                    f"batch dimension. Provided dynamic info for input {i} "
                    f"is: {input_shape}. Please use padding for the other "
                    f"dimensions."
                )

        return True

    def execute(
        self,
        model: torch.nn.Module,
        model_params: ModelParams,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Optimize the input model using pytorch built-in techniques.

        Args:
            model (torch.nn.Module): The pytorch model.
            model_params (ModelParams): The model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction. Default: None.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with a higher drop
                will be ignored. Default: None.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used. Default: None.
            input_data (DataManager): User defined data. Default: None.
        """

        if quantization_type not in self.supported_ops[self.device.type.value]:
            self.compiled_model = None
            return

        if quantization_type is QuantizationType.STATIC and input_data is None:
            raise ValueError("Input data is required for static quantization.")

        self.logger.info(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )

        check_quantization(quantization_type, metric_drop_ths)
        dynamic_batch_size = self._check_dynamic_shape(model_params)

        self.compiled_model = self._compile_model(
            model,
            input_data,
            quantization_type,
            dynamic_batch_size=dynamic_batch_size,
        )

    @torch.no_grad()
    def _compile_model(
        self,
        model: torch.nn.Module,
        input_data: DataManager,
        quantization_type: QuantizationType,
        dynamic_batch_size: bool,
    ) -> torch.jit.ScriptModule:
        input_sample = input_data.get_list(1)[0]
        if self.device.type is DeviceType.GPU:
            if quantization_type is QuantizationType.HALF:
                input_sample = [
                    t.to(self.device.to_torch_format()).half()
                    if torch.is_floating_point(t)
                    else t.to(self.device.to_torch_format())
                    for t in input_sample
                ]
            else:
                input_sample = [
                    t.to(self.device.to_torch_format()) for t in input_sample
                ]
            model.to(self.device.to_torch_format())
        model.eval()

        try:
            model_scripted = symbolic_trace(model)
            model_scripted = torch_neuron.trace(
                model_scripted,
                input_sample,
                dynamic_batch_size=dynamic_batch_size,
                compiler_args=["--fast-math", "none"]
                if quantization_type is None
                else None,
            )
        except Exception:
            try:
                model_scripted = torch_neuron.trace(
                    model,
                    input_sample,
                    dynamic_batch_size=dynamic_batch_size,
                    compiler_args=["--fast-math", "none"]
                    if quantization_type is None
                    else None,
                )
            except Exception:
                raise RuntimeError("Unable to trace model with torch_neuron.")

        return model_scripted

    @torch.no_grad()
    def _quantize_model(
        self,
        model: torch.nn.Module,
        quantization_type: QuantizationType,
        input_tfms: MultiStageTransformation,
        input_data_torch: List[Tuple[torch.Tensor, ...]],
    ):
        raise NotImplementedError()



from nebullvm.optional_modules.dummy import DummyClass

try:
    from transformers import PreTrainedModel, CLIPTextModel, CLIPTokenizer
    from transformers.tokenization_utils import PreTrainedTokenizer
    from transformers.models.bert.modeling_bert import (
        BertModel,
        BertEmbeddings,
        BertEncoder,
        BertPooler,
        BertPreTrainedModel,
    )
    from transformers import BertConfig, GPT2Tokenizer, GPT2LMHeadModel
except ImportError:
    # add placeholders for function definition
    PreTrainedModel = DummyClass
    CLIPTextModel = DummyClass
    CLIPTokenizer = DummyClass
    PreTrainedTokenizer = DummyClass
    BertModel = DummyClass
    BertEmbeddings = DummyClass
    BertEncoder = DummyClass
    BertPooler = DummyClass
    BertPreTrainedModel = DummyClass
    BertConfig = DummyClass
    GPT2Tokenizer = DummyClass
    GPT2LMHeadModel = DummyClass



from nebullvm.operations.inference_learners.torchscript import (
    TorchScriptInferenceLearner,
)


class FasterTransformerInferenceLearner(TorchScriptInferenceLearner):
    MODEL_NAME = "faster_transformer_model_scripted.pt"
    name = "FasterTransformer"



from pathlib import Path
from typing import Union

from nebullvm.operations.inference_learners.torchscript import (
    TorchScriptInferenceLearner,
)


class TorchDynamoInferenceLearner(TorchScriptInferenceLearner):
    name = "TorchDynamo"

    def save(self, path: Union[str, Path], **kwargs):
        # TODO: Implement save function
        # Saving it like a normal PyTorch model raises this error:
        # https://github.com/pytorch/pytorch/issues/93470
        raise NotImplementedError

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        # TODO: Implement load function
        raise NotImplementedError



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Iterable, Callable, List, Union, Dict, Optional

from nebullvm.config import TRAIN_TEST_SPLIT_RATIO
from nebullvm.core import types
from nebullvm.core.models import (
    OptimizeInferenceResult,
    OriginalModel,
    OptimizedModel,
    BenchmarkOriginalModelResult,
    ModelCompiler,
    ModelCompressor,
    OptimizationTime,
    ModelParams,
    DeepLearningFramework,
)
from nebullvm.operations.base import Operation
from nebullvm.operations.conversions.utils import get_conversion_op
from nebullvm.operations.measures.measures import LatencyOriginalModelMeasure
from nebullvm.operations.measures.utils import QUANTIZATION_METRIC_MAP
from nebullvm.operations.optimizations.optimizers.optimizers import (
    PytorchOptimizer,
    TensorflowOptimizer,
    ONNXOptimizer,
)
from nebullvm.operations.optimizations.utils import (
    map_compilers_and_compressors,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import DataLoader as TorchDataLoader
from nebullvm.optional_modules.torch import torch
from nebullvm.optional_modules.utils import (
    check_dependencies,
)
from nebullvm.tools.adapters import (
    ModelAdapter,
    DiffusionAdapter,
    HuggingFaceAdapter,
)
from nebullvm.tools.data import DataManager
from nebullvm.tools.diffusers import (
    is_diffusion_model_pipe,
    is_diffusion_model,
)
from nebullvm.tools.hardware_utils import get_hw_setup
from nebullvm.tools.utils import (
    is_huggingface_data,
    check_input_data,
    is_data_subscriptable,
    get_dl_framework,
    extract_info_from_data,
    get_model_name,
    get_model_size_mb,
    get_throughput,
)


class OptimizeInferenceOp(Operation):
    def __init__(self):
        super().__init__()
        self.torch_optimization_op = PytorchOptimizer()
        self.onnx_optimization_op = ONNXOptimizer()
        self.tensorflow_optimization_op = TensorflowOptimizer()

    @staticmethod
    def _as_data_manager(data) -> DataManager:
        if isinstance(data, DataManager):
            return data
        if check_input_data(data) is False:
            raise ValueError(
                "The provided data does not match the expected "
                "format.\n"
                "Speedster supports data in the following formats: \n"
                "- PyTorch DataLoader\n"
                "- TensorFlow Dataset\n"
                "- List of tuples: [((input_0, ... ), label), ...] \n"
                "Inputs and labels should be either tensors or numpy "
                "arrays,\n"
                "depending on the framework used.\n"
            )
        if is_data_subscriptable(data):
            return DataManager(data)
        else:
            return DataManager.from_iterable(data)

    @staticmethod
    def _check_inputs(model: Any, input_data: types.InputData):
        if model is None:
            raise ValueError("Input model cannot be None")
        if len(input_data) == 0:
            raise ValueError("Input data cannot be empty")

    def execute(
        self,
        model: Any,
        input_data: types.InputData,
        metric_drop_ths: float = None,
        metric: Union[str, Callable] = None,
        optimization_time: str = "constrained",
        dynamic_info: Dict = None,
        config_file: str = None,
        ignore_compilers: List[str] = None,
        ignore_compressors: List[str] = None,
        store_latencies: bool = False,
        **kwargs,
    ) -> OptimizeInferenceResult:

        self._check_inputs(model, input_data)
        check_dependencies(self.device)

        ignore_compilers = map_compilers_and_compressors(
            ignore_compilers, ModelCompiler
        )
        ignore_compressors = map_compilers_and_compressors(
            ignore_compressors, ModelCompressor
        )

        optimization_time = OptimizationTime(optimization_time)

        data = input_data

        if isinstance(data, (TorchDataLoader, tf.data.Dataset)):
            try:
                data = DataManager.from_dataloader(data)
            except Exception:
                raise ValueError(
                    "The provided dataloader does not match the expected "
                    "format.\n"
                    "Speedster supports dataloaders that return tuples in "
                    "the\n"
                    "following formats: \n"
                    "Single input: (input,  label)\n"
                    "Multiple inputs: ((input1, input2, ...),  label) or "
                    "(input1, input2, ...,  label)\n"
                    "Inputs and labels should be either tensors or numpy "
                    "arrays,\n"
                    "depending on the framework used.\n"
                )

        # Setup adapters
        model_adapter: Optional[ModelAdapter] = None
        if is_diffusion_model_pipe(model):
            self.logger.info(
                "The provided model is a diffusion model. "
                "Speedster will optimize the UNet part of the model."
            )
            model_adapter = DiffusionAdapter(model, data, self.device)
        elif is_huggingface_data(data[0]):
            model_adapter = HuggingFaceAdapter(
                model, data, self.device, **kwargs
            )
            if dynamic_info is None:
                self.logger.warning(
                    "Dynamic shape info has not been provided for the "
                    "HuggingFace model. The resulting optimized model "
                    "will be usable only with a fixed input shape. "
                    "To optimize the model for dynamic shapes, please "
                    "look here: https://nebuly.gitbook.io/nebuly/modules/"
                    "speedster/how-to-guides"
                    "#using-dynamic-shape."
                )

        # Adapt data and model
        if model_adapter is not None:
            data = model_adapter.adapted_data
            model = model_adapter.adapted_model

        data = self._as_data_manager(data)
        dl_framework = get_dl_framework(model)

        if metric_drop_ths is not None and metric_drop_ths <= 0:
            metric_drop_ths = None
        elif metric_drop_ths is not None and metric is None:
            metric = "numeric_precision"
        if isinstance(metric, str):
            metric = QUANTIZATION_METRIC_MAP.get(metric)

        model_params: ModelParams = extract_info_from_data(
            model=model,
            input_data=data,
            dl_framework=dl_framework,
            dynamic_info=dynamic_info,
            device=self.device,
            is_diffusion=is_diffusion_model(model),
        )

        data.split(TRAIN_TEST_SPLIT_RATIO)

        # -------- Benchmark original model --------
        original_latency_op = LatencyOriginalModelMeasure().to(self.device)
        orig_model_benchmark: BenchmarkOriginalModelResult = (
            original_latency_op.execute(
                model=model,
                input_data=data.get_split("test"),
                dl_framework=dl_framework,
            )
        )
        original_model = OriginalModel(
            model=model,
            latency_seconds=orig_model_benchmark.latency_seconds,
            name=get_model_name(model),
            size_mb=get_model_size_mb(model),
            framework=dl_framework,
            throughput=get_throughput(
                latency=orig_model_benchmark.latency_seconds,
                # Normal models have batch size B, diffusion
                # models have batch size 2B
                batch_size=model_params.batch_size
                if not is_diffusion_model(model)
                else model_params.batch_size / 2,
            ),
        )
        # ------------------------------------------

        with TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir) / "fp32"
            tmp_dir.mkdir(parents=True, exist_ok=True)

            # Convert model to all available frameworks
            conversion_op = get_conversion_op(dl_framework)
            conversion_op.to(self.device).set_state(model, data).execute(
                save_path=tmp_dir,
                model_params=model_params,
            )

            # Optimize models
            optimized_models: List[OptimizedModel] = []
            is_diffusion = is_diffusion_model(model)
            for i, model in enumerate(conversion_op.get_result()):
                optimized_models += self._optimize(
                    model=model,
                    input_data=data,
                    model_outputs=orig_model_benchmark.model_outputs,
                    optimization_time=optimization_time,
                    metric_drop_ths=metric_drop_ths,
                    metric=metric,
                    model_params=model_params,
                    ignore_compilers=ignore_compilers,
                    ignore_compressors=ignore_compressors,
                    source_dl_framework=dl_framework,
                    pipeline_idx=i + 1,
                    len_pipelines=len(conversion_op.get_result()),
                    is_diffusion=is_diffusion,
                )

        optimized_models.sort(key=lambda x: x.latency_seconds, reverse=False)

        # Check if at least one optimized model has been created
        no_optimized_models = len(optimized_models) < 1
        no_inference_learners = all(
            o.inference_learner is None for o in optimized_models
        )
        if no_optimized_models or no_inference_learners:
            self.logger.warning(
                "No optimized model has been created. This is likely "
                "due to a bug during optimization. Please open an issue "
                "and report in details your use case."
            )

        # Extract lowest-latency model
        lowest_latency = self._extract_lowest_latency_model(optimized_models)

        if model_adapter is not None:
            original_model = model_adapter.adapt_original_model(original_model)
            lowest_latency = model_adapter.adapt_inference_learner(
                lowest_latency
            )

        return OptimizeInferenceResult(
            original_model=original_model,
            optimized_model=lowest_latency,
            hardware_setup=get_hw_setup(),
        )

    def _optimize(
        self,
        model: Any,
        model_outputs: Iterable,
        input_data: types.InputData,
        optimization_time: OptimizationTime,
        metric_drop_ths: float,
        metric: Callable,
        model_params: ModelParams,
        ignore_compilers: List[ModelCompiler],
        ignore_compressors: List[ModelCompressor],
        source_dl_framework: DeepLearningFramework,
        pipeline_idx: int,
        len_pipelines: int,
        is_diffusion: bool,
    ) -> List[OptimizedModel]:
        if isinstance(model, torch.nn.Module):
            optimization_op = self.torch_optimization_op
            self.logger.info(
                f"[{pipeline_idx}/{len_pipelines}] Running PyTorch "
                f"Optimization Pipeline"
            )
        elif isinstance(model, tf.Module):
            optimization_op = self.tensorflow_optimization_op
            self.logger.info(
                f"[{pipeline_idx}/{len_pipelines}] Running TensorFlow "
                f"Optimization Pipeline"
            )
        else:
            optimization_op = self.onnx_optimization_op
            self.logger.info(
                f"[{pipeline_idx}/{len_pipelines}] Running ONNX "
                f"Optimization Pipeline"
            )

        # Run optimization
        optimized_models = optimization_op.to(self.device).execute(
            model=model,
            input_data=input_data,
            optimization_time=optimization_time,
            metric_drop_ths=metric_drop_ths,
            metric=metric,
            model_params=model_params,
            model_outputs=model_outputs,
            ignore_compilers=ignore_compilers,
            ignore_compressors=ignore_compressors,
            source_dl_framework=source_dl_framework,
            is_diffusion=is_diffusion,
        )

        if isinstance(model, torch.nn.Module):
            optimization_op.free_model_gpu(model)

        return optimized_models

    @staticmethod
    def _extract_lowest_latency_model(
        models: List[OptimizedModel],
    ) -> Optional[OptimizedModel]:
        # fmt: off
        inference_learner_models = [
            m for m in models
            if m.inference_learner is not None
        ]
        # fmt: on
        if len(inference_learner_models) == 0:
            return None
        return min(inference_learner_models, key=lambda m: m.latency_seconds)



from nebullvm.optional_modules.dummy import DummyClass

try:
    import diffusers  # noqa F401
    from diffusers import (
        StableDiffusionPipeline,
        DiffusionPipeline,
    )  # noqa F401
    from diffusers.models import (
        AutoencoderKL,
        UNet2DConditionModel,
    )  # noqa F401
    from diffusers.models.unet_2d import UNet2DOutput  # noqa F401
except ImportError:
    diffusers = DummyClass
    StableDiffusionPipeline = DummyClass
    DiffusionPipeline = DummyClass
    UNet2DConditionModel = DummyClass
    AutoencoderKL = DummyClass
    UNet2DOutput = DummyClass

try:
    import onnx_graphsurgeon  # noqa F401
except ImportError:
    onnx_graphsurgeon = DummyClass



# The torch import is necessary for a strange issue when
# using cuda 11.8, if torch is imported after
# tensorflow it generates a core dumped error
from nebullvm.optional_modules.torch import torch  # noqa F401
from nebullvm.tools.logger import setup_logger

setup_logger()

__all__ = [k for k in globals().keys() if not k.startswith("_")]



import os
from pathlib import Path
from tempfile import TemporaryDirectory

from nebullvm.operations.inference_learners.torchscript import (
    TorchScriptInferenceLearner,
)


class TorchNeuronInferenceLearner(TorchScriptInferenceLearner):
    name = "TorchNeuron"

    def get_size(self):
        with TemporaryDirectory() as tmp_dir:
            self.save(tmp_dir)
            return sum(
                os.path.getsize(Path(tmp_dir) / f)
                for f in os.listdir(Path(tmp_dir))
                if os.path.isfile(Path(tmp_dir) / f)
            )



# Based on https://github.com/NVIDIA/FasterTransformer/blob/4402759e48f2340220638675f464b6ba1f79ac3c/examples/pytorch/gpt/utils/huggingface_gpt_convert.py # noqa: E501
# Copyright (c) 2021-2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Convert huggingface GPT model. Use https://huggingface.co/gpt2 as demo.
"""

import argparse
import configparser
import os
import sys

from loguru import logger
import numpy as np
from transformers import GPT2Model  # transformers-4.10.0-py3

from nebullvm.optional_modules.torch import torch

dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(dir_path + "/../../../..")
sys.path.append(dir_path)


def get_weight_data_type(data_type):
    if data_type == "fp32":
        return np.float32
    elif data_type == "fp16":
        return np.float16
    else:
        assert False, f"Invalid weight data type {data_type}"


def split_and_convert_process(i, saved_dir, factor, key, args, val):

    if (
        key.find("input_layernorm.weight") != -1
        or key.find("input_layernorm.bias") != -1
        or key.find("attention.dense.bias") != -1
        or key.find("post_attention_layernorm.weight") != -1
        or key.find("post_attention_layernorm.bias") != -1
        or key.find("mlp.dense_4h_to_h.bias") != -1
        or key.find("final_layernorm.weight") != -1
        or key.find("final_layernorm.bias") != -1
    ):

        # shared weights, only need to convert the weights of rank 0
        if i == 0:
            saved_path = saved_dir + "/model." + key + ".bin"
            val.tofile(saved_path)

    elif (
        key.find("attention.dense.weight") != -1
        or key.find("mlp.dense_4h_to_h.weight") != -1
    ):
        split_vals = np.split(val, factor, axis=0)
        for j in range(factor):
            saved_path = (
                saved_dir + "/model." + key + ".%d.bin" % (i * factor + j)
            )
            split_vals[j].tofile(saved_path)

    elif (
        key.find("mlp.dense_h_to_4h.weight") != -1
        or key.find("mlp.dense_h_to_4h.bias") != -1
    ):

        split_vals = np.split(val, factor, axis=-1)
        for j in range(factor):
            saved_path = (
                saved_dir + "/model." + key + ".%d.bin" % (i * factor + j)
            )
            split_vals[j].tofile(saved_path)

    elif key.find("attention.query_key_value.bias") != -1:
        local_dim = (int)(val.shape[-1] / 3)

        val = val.reshape(3, local_dim)
        split_vals = np.split(val, factor, axis=-1)

        for j in range(factor):
            saved_path = (
                saved_dir + "/model." + key + ".%d.bin" % (i * factor + j)
            )
            split_vals[j].tofile(saved_path)

    elif key.find("attention.query_key_value.weight") != -1:
        hidden_dim = val.shape[0]
        local_dim = (int)(val.shape[-1] / 3)

        val = val.reshape(hidden_dim, 3, local_dim)
        split_vals = np.split(val, factor, axis=-1)

        for j in range(factor):
            saved_path = (
                saved_dir + "/model." + key + ".%d.bin" % (i * factor + j)
            )
            split_vals[j].tofile(saved_path)

    else:
        logger.warning("[ERROR] cannot find key '{}'".format(key))


def split_and_convert(args):
    torch_device = "cuda" if torch.cuda.is_available() else "cpu"
    model = GPT2Model.from_pretrained(args.in_file).to(torch_device)
    main(
        args.saved_dir,
        model,
        args.trained_gpu_num,
        args.infer_gpu_num,
        args.processes,
        args.weight_data_type,
    )


def main(
    saved_dir,
    model: GPT2Model,
    trained_gpu_num=1,
    infer_gpu_num=1,
    processes=1,
    weight_data_type="fp32",
):
    assert isinstance(model, GPT2Model), "model must be GPT2Model"
    args = None
    saved_dir = saved_dir + "/%d-gpu/" % infer_gpu_num

    if not os.path.exists(saved_dir):
        os.makedirs(saved_dir)
    # ckpt_name = args.in_file

    t_gpu_num = trained_gpu_num
    i_gpu_num = infer_gpu_num
    assert i_gpu_num % t_gpu_num == 0

    factor = (int)(i_gpu_num / t_gpu_num)

    # load position_embedding from rank 0
    # torch_device = 'cuda' if torch.cuda.is_available() else 'cpu'
    # model = GPT2Model.from_pretrained(args.in_file).to(torch_device)

    hf_config = vars(model.config)

    # NOTE: save parameters to config files (loaded by triton backends)
    config = configparser.ConfigParser()
    config["gpt"] = {}
    try:
        config["gpt"]["model_name"] = (
            "gpt"
            if hf_config["_name_or_path"] == ""
            else hf_config["_name_or_path"]
        )
        config["gpt"]["head_num"] = str(hf_config["n_head"])
        n_embd = hf_config["n_embd"]
        config["gpt"]["size_per_head"] = str(n_embd // hf_config["n_head"])
        config["gpt"]["inter_size"] = str(n_embd * 4)
        config["gpt"]["max_pos_seq_len"] = str(hf_config["n_positions"])
        config["gpt"]["num_layer"] = str(hf_config["n_layer"])
        config["gpt"]["vocab_size"] = str(hf_config["vocab_size"])
        config["gpt"]["start_id"] = str(hf_config["bos_token_id"])
        config["gpt"]["end_id"] = str(hf_config["eos_token_id"])
        config["gpt"]["weight_data_type"] = weight_data_type
        with open(saved_dir + "/config.ini", "w") as configfile:
            config.write(configfile)
    except:  # noqa: E722
        logger.warning("Fail to save the config in config.ini.")

    np_weight_data_type = get_weight_data_type(weight_data_type)

    huggingface_model_name_pattern = [
        "ln_1.bias",
        "ln_1.weight",
        "attn.c_attn.bias",
        "attn.c_attn.weight",
        "attn.c_proj.bias",
        "attn.c_proj.weight",
        "ln_2.bias",
        "ln_2.weight",
        "mlp.c_fc.bias",
        "mlp.c_fc.weight",
        "mlp.c_proj.bias",
        "mlp.c_proj.weight",
    ]

    ft_model_name_pattern = [
        "input_layernorm.bias",
        "input_layernorm.weight",
        "attention.query_key_value.bias",
        "attention.query_key_value.weight",
        "attention.dense.bias",
        "attention.dense.weight",
        "post_attention_layernorm.bias",
        "post_attention_layernorm.weight",
        "mlp.dense_h_to_4h.bias",
        "mlp.dense_h_to_4h.weight",
        "mlp.dense_4h_to_h.bias",
        "mlp.dense_4h_to_h.weight",
    ]

    # torch.multiprocessing.set_start_method("spawn")
    # torch.multiprocessing.set_sharing_strategy("file_system")
    # pool = multiprocessing.Pool(args.processes)
    for name, param in model.named_parameters():
        if name.find("weight") == -1 and name.find("bias") == -1:
            continue
        if name == "wpe.weight":
            param.detach().cpu().numpy().astype(np_weight_data_type).tofile(
                saved_dir + "model.wpe.bin"
            )
        elif name == "wte.weight":
            param.detach().cpu().numpy().astype(np_weight_data_type).tofile(
                saved_dir + "model.wte.bin"
            )
        elif name == "ln_f.bias":
            param.detach().cpu().numpy().astype(np_weight_data_type).tofile(
                saved_dir + "model.final_layernorm.bias.bin"
            )
        elif name == "ln_f.weight":
            param.detach().cpu().numpy().astype(np_weight_data_type).tofile(
                saved_dir + "model.final_layernorm.weight.bin"
            )
        elif name == "lm_head.weight":
            param.detach().cpu().numpy().astype(np_weight_data_type).tofile(
                saved_dir + "model.lm_head.weight.bin"
            )
        else:
            for i in range(len(huggingface_model_name_pattern)):
                if name.find(huggingface_model_name_pattern[i]) != -1:
                    new_name = name.replace("h.", "layers.").replace(
                        huggingface_model_name_pattern[i],
                        ft_model_name_pattern[i],
                    )
                    # pool.starmap(split_and_convert_process,
                    # [(0, saved_dir, factor, new_name, args,
                    # param.detach().cpu().numpy().astype(np_weight_data_type))],
                    # )
                    split_and_convert_process(
                        0,
                        saved_dir,
                        factor,
                        new_name,
                        args,
                        param.detach()
                        .cpu()
                        .numpy()
                        .astype(np_weight_data_type),
                    )

    # pool.close()
    # pool.join()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "-saved_dir",
        "-o",
        type=str,
        help="file name of output file",
        required=True,
    )
    parser.add_argument(
        "-in_file",
        "-i",
        type=str,
        help="file name of input checkpoint file",
        required=True,
    )
    parser.add_argument(
        "-trained_gpu_num",
        "-t_g",
        type=int,
        help="How many gpus for inference",
        default=1,
    )
    parser.add_argument(
        "-infer_gpu_num",
        "-i_g",
        type=int,
        help="How many gpus for inference",
        required=True,
    )
    parser.add_argument(
        "-processes",
        "-p",
        type=int,
        help="How many processes to spawn for conversion (default: 4)",
        default=4,
    )
    parser.add_argument(
        "-weight_data_type", type=str, default="fp32", choices=["fp32", "fp16"]
    )

    args = parser.parse_args()
    logger.info("\n=============== Argument ===============")
    for key in vars(args):
        logger.info("{}: {}".format(key, vars(args)[key]))
    logger.info("========================================")

    split_and_convert(args)



from nebullvm.optional_modules.dummy import DummyClass

try:
    import tensorrt
    from tensorrt import IInt8EntropyCalibrator2
except ImportError:
    tensorrt = DummyClass
    IInt8EntropyCalibrator2 = DummyClass

try:
    import polygraphy.cuda as polygraphy
    from polygraphy.logger import G_LOGGER

    G_LOGGER.module_severity = 40
    from polygraphy.backend.onnx.loader import fold_constants
except ImportError:
    polygraphy = DummyClass
    fold_constants = DummyClass



from nebullvm.optional_modules.dummy import DummyClass

try:
    import onnxruntime  # noqa F401
    from onnxruntime.quantization import (
        QuantType,
        quantize_static,
        quantize_dynamic,
        CalibrationDataReader,
    )
except ImportError:
    onnxruntime = DummyClass
    setattr(onnxruntime, "SessionOptions", None)
    QuantType = quantize_static = quantize_dynamic = None
    CalibrationDataReader = DummyClass
except FileNotFoundError:
    # Solves a colab issue
    QuantType = quantize_static = quantize_dynamic = None
    CalibrationDataReader = DummyClass

try:
    # They require torch
    from onnxruntime.transformers import optimizer
    from onnxruntime.transformers.optimizer import MODEL_TYPES
except ImportError:
    MODEL_TYPES = DummyClass
    optimizer = DummyClass



# Based on: https://github.com/NVIDIA/FasterTransformer/blob/4402759e48f2340220638675f464b6ba1f79ac3c/examples/pytorch/bert/utils/checkpoint_quantization.py # noqa: E501
# Copyright (c) 2020-2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import numpy as np
from loguru import logger

from nebullvm.optional_modules.torch import torch

ACTIVATION_AMAX_NUM = 72
INT8O_GEMM_NUM = 8
TRT_FUSED_MHA_AMAX_NUM = 3
SCALE_RESERVE_NUM = 21


def checkpoint_quantization(
    init_dict, sparse, ths_path="./lib/libth_transformer.so"
):
    logger.info("Quantizing checkpoint ...")
    torch.classes.load_library(ths_path)
    weight_quantize = torch.ops.fastertransformer.weight_quantize

    def init_graph():
        layer_num = 0
        regex = re.compile("layer.\d+")  # noqa: W605
        amaxTotalNum = 0
        for name, tensor_value in init_dict.items():
            if "intermediate.dense.weight" in name and amaxTotalNum == 0:
                amaxTotalNum = (
                    ACTIVATION_AMAX_NUM
                    + 9 * tensor_value.size(1)
                    + INT8O_GEMM_NUM
                    + TRT_FUSED_MHA_AMAX_NUM
                    + SCALE_RESERVE_NUM
                )
            tmp = regex.findall(name)
            if len(tmp) < 1:
                continue
            num_tmp = int(tmp[0].replace("layer.", ""))
            if layer_num < num_tmp:
                layer_num = num_tmp
        layer_num = layer_num + 1
        # add new var for amax
        for i in range(layer_num):
            init_dict[
                "bert.encoder.layer.{}.amaxList".format(i)
            ] = torch.zeros((amaxTotalNum,), dtype=torch.float32)
        return layer_num, amaxTotalNum

    layer_num, amaxTotalNum = init_graph()

    kernel_name_list = [
        "attention.self.query",
        "attention.self.key",
        "attention.self.value",
        "attention.output.dense",
        "intermediate.dense",
        "output.dense",
    ]

    amax_name_list = [
        "attention.self.query._input_quantizer",
        "attention.self.query._aftergemm_quantizer",
        "attention.self.matmul_q_input_quantizer",
        "attention.self.key._aftergemm_quantizer",
        "attention.self.matmul_k_input_quantizer",
        "attention.self.value._aftergemm_quantizer",
        "attention.self.matmul_v_input_quantizer",
        "attention.self.softmax_input_quantizer",
        "attention.self.matmul_a_input_quantizer",
        "attention.output.dense._input_quantizer",
        "attention.output.dense._aftergemm_quantizer",
        "intermediate.dense._input_quantizer",
        "intermediate.dense._aftergemm_quantizer",
        "output.dense._input_quantizer",
        "output.dense._aftergemm_quantizer",
        "special_F2Bias_scale",
    ]

    int8O_gemm_weight_amax_list = [0 for i in range(INT8O_GEMM_NUM)]
    int8O_gemm_weight_list = [
        "attention.self.query",
        "attention.self.key",
        "attention.self.value",
        "attention.self.matmul_k_input_quantizer",
        "attention.self.matmul_v_input_quantizer",
        "attention.output.dense",
        "intermediate.dense",
        "output.dense",
    ]

    int8O_gemm_input_amax_list = [0 for i in range(INT8O_GEMM_NUM)]
    int8O_gemm_input_list = [
        "attention.self.query._input_quantizer",
        "attention.self.key._input_quantizer",
        "attention.self.value._input_quantizer",
        "attention.self.matmul_q_input_quantizer",
        "attention.self.matmul_a_input_quantizer",
        "attention.output.dense._input_quantizer",
        "intermediate.dense._input_quantizer",
        "output.dense._input_quantizer",
    ]

    int8O_gemm_output_amax_list = [0 for i in range(INT8O_GEMM_NUM)]
    int8O_gemm_output_list = [
        "attention.self.query._aftergemm_quantizer",
        "attention.self.key._aftergemm_quantizer",
        "attention.self.value._aftergemm_quantizer",
        "attention.self.softmax_input_quantizer",
        "attention.output.dense._input_quantizer",
        "attention.output.dense._aftergemm_quantizer",
        "intermediate.dense._aftergemm_quantizer",
        "output.dense._aftergemm_quantizer",
    ]

    same_value_tuple_list = [
        (
            "attention.self.query._input_quantizer",
            "attention.self.key._input_quantizer",
            "attention.self.value._input_quantizer",
            "attention.output.add_residual_input_quantizer",
        ),
        (
            "intermediate.dense._input_quantizer",
            "output.add_residual_input_quantizer",
        ),
    ]

    factor = 1000000.0  # noqa: F841
    for i in range(layer_num):
        amaxList = np.zeros([amaxTotalNum]).astype(np.float32)
        amax_id = 0
        # verify some quantizers have same value.
        # input_quantizer is per-tensor quantization
        for same_value_tuple in same_value_tuple_list:
            tmp_v = init_dict[
                "bert.encoder.layer.{}.{}._amax".format(i, same_value_tuple[0])
            ].numpy()
            for same_value_name in same_value_tuple:
                tmp_v_2 = init_dict[
                    "bert.encoder.layer.{}.{}._amax".format(i, same_value_name)
                ].numpy()
                assert np.allclose(tmp_v, tmp_v_2)

        for amax_name in amax_name_list:
            if amax_name == "special_F2Bias_scale":
                if i != layer_num - 1:
                    quant_max = init_dict[
                        "bert.encoder.layer.{}.{}._amax".format(
                            i + 1, amax_name_list[0]
                        )
                    ].item()
                    amax = abs(quant_max)
                else:
                    # not used, placeholder
                    amax = 1.0
                amaxList[amax_id] = amax
                amax_id += 1
                amaxList[amax_id] = amax / 127.0
                amax_id += 1
                amaxList[amax_id] = amax / 127.0 / 127.0
                amax_id += 1
                amaxList[amax_id] = 127.0 / amax
                amax_id += 1
                continue

            quant_max = init_dict[
                "bert.encoder.layer.{}.{}._amax".format(i, amax_name)
            ].item()
            amax = abs(quant_max)  # round(abs(quant_max)*factor)/factor
            if amax_name in int8O_gemm_input_list:
                int8O_gemm_input_amax_list[
                    int8O_gemm_input_list.index(amax_name)
                ] = amax
                if amax_name == "attention.self.query._input_quantizer":
                    int8O_gemm_input_amax_list[
                        int8O_gemm_input_list.index(
                            "attention.self.key._input_quantizer"
                        )
                    ] = amax
                    int8O_gemm_input_amax_list[
                        int8O_gemm_input_list.index(
                            "attention.self.value._input_quantizer"
                        )
                    ] = amax
            if amax_name in int8O_gemm_output_list:
                int8O_gemm_output_amax_list[
                    int8O_gemm_output_list.index(amax_name)
                ] = amax
            if amax_name in int8O_gemm_weight_list:
                int8O_gemm_weight_amax_list[
                    int8O_gemm_weight_list.index(amax_name)
                ] = amax
            amaxList[amax_id] = amax
            amax_id += 1
            amaxList[amax_id] = amax / 127.0
            amax_id += 1
            amaxList[amax_id] = amax / 127.0 / 127.0
            amax_id += 1
            amaxList[amax_id] = 127.0 / amax
            amax_id += 1

        # kernel amax starts from ACTIVATION_AMAX_NUM
        assert amax_id == 64
        amax_id = ACTIVATION_AMAX_NUM
        for kernel_id, kernel_name in enumerate(kernel_name_list):
            kernel = (
                init_dict[
                    "bert.encoder.layer.{}.{}.weight".format(i, kernel_name)
                ]
                .transpose(-1, -2)
                .contiguous()
            )
            quant_max2 = init_dict[
                "bert.encoder.layer.{}.{}._weight_quantizer._amax".format(
                    i, kernel_name
                )
            ]
            amax2 = abs(quant_max2)
            if amax2.dim() == 0:
                quant_max_processed = torch.full(
                    (kernel.size(1),),
                    amax2.item(),
                    dtype=amax2.dtype,
                    device=amax2.device,
                )
            else:
                quant_max_processed = amax2.view(-1)
            kernel_processed = weight_quantize(
                kernel, quant_max_processed.cuda(), sparse
            )
            init_dict[
                "bert.encoder.layer.{}.{}.weight".format(i, kernel_name)
            ] = kernel_processed
            if kernel_name in int8O_gemm_weight_list:
                int8O_gemm_weight_amax_list[
                    int8O_gemm_weight_list.index(kernel_name)
                ] = quant_max_processed[0]
            for e in quant_max_processed:
                amaxList[amax_id] = e
                amax_id += 1

        # for int8O gemm deQuant
        for j in range(INT8O_GEMM_NUM):
            amaxList[amax_id] = (
                int8O_gemm_input_amax_list[j] * int8O_gemm_weight_amax_list[j]
            ) / (127.0 * int8O_gemm_output_amax_list[j])
            amax_id += 1

        # for trt fused MHA amax
        # QKV_addBias_amax
        amaxList[amax_id] = np.maximum(
            np.maximum(amaxList[8], amaxList[16]), amaxList[24]
        )
        amax_id += 1
        # softmax amax
        amaxList[amax_id] = amaxList[32]
        amax_id += 1
        # bmm2 amax
        amaxList[amax_id] = amaxList[36]
        amax_id += 1

        init_dict["bert.encoder.layer.{}.amaxList".format(i)] = torch.tensor(
            amaxList, dtype=torch.float32
        )
    logger.info("Quantizing checkpoint done.")
    return init_dict
